
#include <Lib/Printer.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Arguments.h>
#include <Lib/Debug.h>
#include <Lib/fmanip.h>
#include <Lib/FileWriter.h>
#include <Lib/GZipWriter.h>

#include <Lib/ItemML.h>
#include <Lib/BaseML.h>

class AIDA_ML : public Lib::BaseML {
public:
  AIDA_ML(Lib::ItemMLFactory& aFactory,
          int aVerbose,
          Slash::Core::IWriter& aPrinter)
  :Lib::BaseML(aFactory,aVerbose)
  ,fPrinter(aPrinter),fCompress(false)
  {
  }
public:
  virtual bool visitEndElement(Lib::ItemML& aItemML,bool& aKeep) {
    const std::string& tagName = aItemML.tagName();
    if(tagName=="aida") {
      aKeep = false;
    } else if(isAIDA_Object(tagName)) {
      std::string path;
      if(aItemML.attributeValue("","path",path)) {
        std::string name;
        if(aItemML.attributeValue("","name",name)) write(path,name);
      }  
      aKeep = false;
    } else {
      aKeep = true;
    }
    return true;
  }
  bool write(const std::string& aPath,const std::string& aName) const { 
    std::string fileName = "./"+aPath+"/"+aName+".aida";
    printf("debug : write %s\n",fileName.c_str());
    std::string fullName = Lib::fmanip::name(fileName);
    Slash::Core::IWriter* writer = 0;
    if(fCompress) {
      writer = new Lib::GZipWriter(fullName,fPrinter);
      if(!writer->enabled()) {
        printf("debug : problem\n");
        delete writer;
        return false;
      }
    } else {
      writer = new Lib::FileWriter(fullName,fPrinter);
      if(!writer->enabled()) {
        printf("debug : problem\n");
        delete writer;
        return false;
      }
    }
    Lib::Out out(*writer);

    // Header :
    out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << Lib::endl;
    out << "<!DOCTYPE aida SYSTEM \"http://aida.freehep.org/schemas/3.0/aida.dtd\">" << Lib::endl;
    //out << "<aida version=" << Lib::sout(AIDA_VERSION) << ">" << Lib::endl;
    //out << "  <implementation package=" << Lib::sout("BatchLab") 
    //    << " version=" << Lib::sout(BATCHLAB_VERSION) << "/>" << Lib::endl;

    //XML_StoreVisitor visitor(fManager,*this,out,printer());
    //rootFolder().visit(visitor);

    // Footer :
    out << "</aida>" << Lib::endl;

    delete writer;
    return true;
  }
  bool isAIDA_Object(const std::string& aTagName) const {
    if(aTagName=="histogram1d") return true;
    if(aTagName=="histogram2d") return true;
    if(aTagName=="histogram3d") return true;
    if(aTagName=="profile1d") return true;
    if(aTagName=="profile2d") return true;
    if(aTagName=="cloud1d") return true;
    if(aTagName=="cloud2d") return true;
    if(aTagName=="cloud3d") return true;
    if(aTagName=="tuple") return true;
    if(aTagName=="dataPointSet") return true;
    if(aTagName=="function") return true;
    return false;
  }
private:
  Slash::Core::IWriter& fPrinter;
  bool fCompress;
};


//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;{

  Lib::Arguments args(aArgc,aArgv);

  Lib::Out out(printer);

  if(args.isAnArgument("-h")) {
    out << "OS> Lib_xml.exe [options] <file>" << Lib::endl;
    out << " -h : help" << Lib::endl;
    return 0;
  }

  int verboseLevel = 0;
  std::string sverbose;
  args.find("-verbose",verboseLevel);

  std::string file;
  if(!args.file(file)) {
    out << "no XML aida file given." << Lib::endl;
    return 0;
  }

  Lib::ItemMLFactory factory;
  AIDA_ML aidaML(factory,verboseLevel,printer);
  std::vector<std::string> tags;
  tags.push_back("aida");
  tags.push_back("histogram1d");
  tags.push_back("histogram2d");
  tags.push_back("histogram3d");
  tags.push_back("profile1d");
  tags.push_back("profile2d");
  tags.push_back("axis");
  tags.push_back("statistics");
  tags.push_back("data1d");
  tags.push_back("data2d");
  tags.push_back("data3d");
  tags.push_back("tuple");
  tags.push_back("columns");
  tags.push_back("rows");
  tags.push_back("row");
  tags.push_back("entryTuple");
  tags.push_back("cloud1d");
  tags.push_back("cloud2d");
  tags.push_back("cloud3d");
  tags.push_back("entries1d");
  tags.push_back("entries2d");
  tags.push_back("entries3d");
  tags.push_back("dataPointSet");
  tags.push_back("dataPoint");
  tags.push_back("function");
  tags.push_back("arguments");
  tags.push_back("argument");
  tags.push_back("parameters");
  tags.push_back("annotation");
  aidaML.setTags(tags);
  if(!aidaML.loadFile(file,false,printer)) {
    out << "Lib_xml : " 
        << " problem when loading file " << Lib::sout(file) << Lib::endl;
  }


 }Lib::Debug::balance(printer);

  return 0;
}
