
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud2D.h>
#include <AIDA/ICloud3D.h>

#include <Lib/Arguments.h>
#include <Lib/Randoms.h>
#include <Lib/Printer.h>
#include <Lib/Debug.h>
#include <Lib/smanip.h>
#include <Lib/Time.h>

#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Time begin(Lib::Time::now());

  int verboseLevel = 0;

 {//  Put everything in a block so that temporary objects
  // on the stack are deleted before the final memory decount is done. 

  Lib::Arguments args(aArgc,aArgv);
  bool help = args.isAnArgument("-h");
  if(help) {
    std::cout << "BatchLab_cloud::main :" << std::endl;
    std::cout << "Options : " << std::endl;
    std::cout << "  OS> BatchLab_cloud -h" << std::endl;
    std::cout << "   to have help." << std::endl;
    std::cout << "  OS> BatchLab_cloud -verbose" << std::endl;
    std::cout << "   to switch to verbose mode." << std::endl;
    std::cout << "  OS> BatchLab_cloud -entries=<integer>" << std::endl;
    std::cout << "   give number of entries to fill the cloud." << std::endl;
    std::cout << "   Default is 1000000." << std::endl;
    return 0;
  }

  std::string sverbose;
  if(args.find("-verbose",sverbose)) {
    if(sverbose=="") verboseLevel = 1;
  }

  //int entries = 1000000;
  int entries = 10000000;
  std::string sentries;
  if(args.find("-entries",sentries)) {
    int ientries;
    if(Lib::smanip::toint(sentries,ientries)) entries = ientries;
  }

  AIDA::IAnalysisFactory* analysisFactory = AIDA_createAnalysisFactory();
  if(!analysisFactory) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't get an analysis factory." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " Have an analysis factory." << std::endl;
  }

  AIDA::ITreeFactory* treeFactory = analysisFactory->createTreeFactory();
  if(!treeFactory) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't get a tree factory." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " Have a tree factory." << std::endl;
  }

  // Creating a tree in memory
  AIDA::ITree* memoryTree = treeFactory->create();
  if(!memoryTree) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't get the default in memory tree." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " Have the default in memory tree." << std::endl;
  }

  // Get an histogram factory :
  AIDA::IHistogramFactory* histogramFactory = 
    analysisFactory->createHistogramFactory(*memoryTree);
  if(!histogramFactory) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't get a histogram factory." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " Have a histogram factory." << std::endl;
  }

  ////////////////////////////
  // Create some data :
  ////////////////////////////
  AIDA::ICloud2D* cloud2D = 
    histogramFactory->createCloud2D("cloud2D","Cloud2D");
  if(!cloud2D) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't create a cloud2D." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " \"cloud2D\" cloud created." << std::endl;
  }

  AIDA::ICloud3D* cloud3D = 
    histogramFactory->createCloud3D("cloud3D","Cloud3D");
  if(!cloud3D) {
    std::cout << "BatchLab_cloud::main :"
              << " Can't create a cloud3D." << std::endl;
    return 1;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " \"cloud3D\" cloud created." << std::endl;
  }

  Lib::RandomGauss randomGauss;

 {Lib::Time beginFill = Lib::Time::now();
  for(int count=0;count<entries;count++)
    cloud2D->fill(randomGauss.shoot(),randomGauss.shoot(),1);
  Lib::Time endFill = Lib::Time::elapsed(beginFill);
  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " fill 2D elapsed time : " 
              << endFill.seconds() << " seconds."
              << std::endl;
  }}

 {Lib::Time beginFill(Lib::Time::now());
  for(int count=0;count<entries;count++)
    cloud3D->fill(randomGauss.shoot(),randomGauss.shoot(),randomGauss.shoot(),1);
  Lib::Time endFill = Lib::Time::elapsed(beginFill);
  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " fill 3D elapsed time : " 
              << endFill.seconds() << " seconds."
              << std::endl;
  }}

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :" << std::endl;
    std::cout << " Cloud 2D : " << cloud2D->title() << std::endl;
    std::cout << " entries = " << cloud2D->entries() << std::endl;
    std::cout << " meanX = " << cloud2D->meanX() 
              << " rmsX = " << cloud2D->rmsX()
              << std::endl;
    std::cout << " meanY = " << cloud2D->meanY() 
              << " rmsY = " << cloud2D->rmsY()
              << std::endl;
    std::cout << std::endl;
    std::cout << " Cloud 3D : " << cloud3D->title() << std::endl;
    std::cout << " entries = " << cloud3D->entries() << std::endl;
    std::cout << " meanX = " << cloud3D->meanX() 
              << " rmsX = " << cloud3D->rmsX()
              << std::endl;
    std::cout << " meanY = " << cloud3D->meanY() 
              << " rmsY = " << cloud3D->rmsY()
              << std::endl;
    std::cout << " meanZ = " << cloud3D->meanZ() 
              << " rmsZ = " << cloud3D->rmsZ()
              << std::endl;
  }


  delete histogramFactory;
  delete treeFactory;
  delete analysisFactory;
  }

  // A simple check to see if object creation/destruction balance is correct :
  Lib::Printer printer;
  Lib::Debug::balance(printer);

  Lib::Time end = Lib::Time::elapsed(begin);

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " elapsed time : " 
              << end.seconds() << " seconds."
              << std::endl;
  }

  if(verboseLevel) {
    std::cout << "BatchLab_cloud::main :"
              << " Exiting..." << std::endl;
  }

  return 0;
}

