// this :
#include <BatchLab/Zip/Tree.h>

#include <Slash/Store/IFolderVisitor.h>
#include <Slash/Store/IObject.h>

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/BaseZip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/ItemML.h>
#include <Lib/ZipWriter.h>

#include <AIDA/IManagedObject.h>
#include <AIDA/IConstants.h>

#include <BatchLab/XML/AIDA_ML.h>
#include <BatchLab/Version.h>

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#include <BatchLab/Core/amanip.h>
#else
#endif

namespace BatchLab {

class AIDA_Zip : public Lib::BaseZip {
public:
  AIDA_Zip(BaseTree& aTree)
  :fTree(aTree)
  ,fVerbose(fTree.verboseLevel())
  //,fVerbose(1)
  {}
public:
  virtual bool visitEntry(const std::string& aEntry,
                          unsigned int aSize,
                          const char* aBuffer) {
    if(fVerbose) {
      Lib::Out out(fTree.printer());
      out << "AIDA_Zip::visitEntry :" 
          << " entry " << Lib::sout(aEntry)
          << " (size " << aSize << ")..."
          << Lib::endl;
    }

    Lib::ItemMLFactory factory;
    AIDA_ML aidaML(factory,fTree);
    if(!aidaML.loadBuffer(aSize,aBuffer,fTree.printer()) || aidaML.errors()) {
      Lib::Out out(fTree.printer());
      out << "AIDA_Zip::visitEntry :" 
          << " problem loading entry " << Lib::sout(aEntry)
          << ". Continue anyway."
          << Lib::endl;
      return true;
    }
    if(fVerbose) {
      Lib::Out out(fTree.printer());
      out << "AIDA_Zip::visitEntry :" 
          << " load of " << Lib::sout(aEntry) << " done."
          << Lib::endl;
    }
    return true;
  }
private:
  BaseTree& fTree;
  int fVerbose;
};

}
//////////////////////////////////////////////////////////////////////////////
BatchLab::Zip::Tree::Tree(
 Slash::Core::ISession& aSession
,const std::string& aStoreName
,bool aReadOnly
,bool aCreateNew
,const std::string& aOptions
,Zip::Manager& aManager
,int aVerboseLevel
)
:MemoryTree(aSession,aStoreName) //FIXME : have a name # storeName
,fManager(aManager)
,fStoreName(aStoreName)
,fNameCounter(0)
,fReadOnly(aReadOnly)
,fCreateNew(aCreateNew)
,fCompress(false)
,fIsValid(true)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setVerboseLevel(aVerboseLevel);

  std::string value;
  if(Lib::smanip::option(aOptions,"compress",value)) {
    bool bval;
    if(Lib::smanip::tobool(value,bval)) {
      fCompress = bval;
    }
  }

  //printf("debug : Zip::Tree :<-- \"%s\" \"%s\" %d -->\n",
  // aStoreName.c_str(),aOptions.c_str(),fCompress);

  if(!Lib::BaseZip::available()) {
    Lib::Out out(printer());
    out << "BatchLab::Zip::Tree::Zip::Tree :"
        << " Zip not available. " 
        << " check installation of Lib package "
        << " and Lib::BaseZip in particular. "
        << Lib::endl;
    fCompress = false;
  }
  
  if(fCreateNew) {
    // Handled in the commit.
  } else {
    // Read the existing file :
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::Zip::Tree::Zip::Tree :"
          << " open " << Lib::sout(aStoreName) 
          << " with compression=" 
          << std::string(fCompress?"yes":"no") << "..." << Lib::endl;
    }
    AIDA_Zip aidaZip(*this);
    if(!aidaZip.loadFile(aStoreName,printer())) {
      Lib::Out out(printer());
      out << "BatchLab::Zip::Tree::Zip::Tree :"
          << " problem loading file " << Lib::sout(aStoreName)
          << Lib::endl;
      fIsValid = false;
      return;
    }
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::Zip::Tree::Zip::Tree :"
          << " load of " << Lib::sout(aStoreName) << " done."
          << Lib::endl;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::Zip::Tree::~Tree(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::Zip::Tree::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(BatchLab::Zip::Tree)
  else return MemoryTree::cast(aClass);
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::Zip::Tree::storeName(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fStoreName;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::Zip::Tree::storeType(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return "BatchLab::Zip::Tree";
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::Zip::Tree::isValid(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fIsValid;
}

//////////////////////////////////////////////////////////////////////////////
/// Data specific area ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

namespace BatchLab {
  class Zip_StoreVisitor : public virtual Slash::Store::IFolderVisitor {
  public:
    Zip_StoreVisitor(Zip::Tree& aTree,
                     Lib::ItemMLFactory& aFactory,
                     Lib::ZipWriter& aWriter,
                     Slash::Core::IWriter& aPrinter)
    :fTree(aTree)
    ,fAIDA_ML(aFactory,aTree)
    ,fWriter(aWriter),fPrinter(aPrinter)
    ,fVerboseLevel(fTree.verboseLevel())
    //,fVerboseLevel(1)
    {}
  public:
    virtual bool visit(const Slash::Store::IFolder& aFolder,
                       const Slash::Store::IObject& aObject){
      if(fVerboseLevel) {
        Lib::Out out(fPrinter);
        out << "BatchLab::Zip_StoreVisitor::visit :" 
            << " path : " << Lib::sout(aFolder.path())
            << " name : " << Lib::sout(aObject.name())
            << Lib::endl;
      }

      // Strip the first slash if any.
      std::string path = aFolder.path()+"/"+aObject.name();
      if(path.size()&&(path[0]=='/')) path = path.substr(1,path.size()-1);

      // Append the type/class of object :
      path += ".";
      if(const AIDA::IManagedObject* amo = 
          Lib_ConstCast(aObject,AIDA::IManagedObject)) {
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
        path += amanip::managed_object_type(*amo);
#else
        path += amo->type();
#endif
        Lib::smanip::replace(path,"AIDA::","");
      } else {
        path += aObject.storeClassName();
      }

      if(!fWriter.openEntry(path)) return false;

      // Header :
      Lib::Out out(fWriter);
      out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << Lib::endl;
      out << "<!DOCTYPE aida SYSTEM \"http://aida.freehep.org/schemas/3.0/aida.dtd\">" << Lib::endl;
      out << "<aida version=" << Lib::sout(AIDA_VERSION) << ">" << Lib::endl;
      out << "  <implementation package=" << Lib::sout("BatchLab") 
          << " version=" << Lib::sout(BATCHLAB_VERSION) << "/>" << Lib::endl;

      if(!fAIDA_ML.write(fWriter,aFolder.path(),aObject)) {
        //continue anyway.
      }

      // Footer :
      out << "</aida>" << Lib::endl;

      return fWriter.closeEntry();
    }
    virtual bool beginVisit(const Slash::Store::IFolder&,
                            const Slash::Store::IFolder& aFolder){ 
      if(fVerboseLevel) {
        Lib::Out out(fPrinter);
        out << "BatchLab::Zip_StoreVisitor::beginVisit :" 
            << " path : " << Lib::sout(aFolder.path())
            << Lib::endl;
      }
      // Strip the first slash if any and append one if needed.
      std::string path = aFolder.path();
      if(path.size()&&(path[0]=='/')) path = path.substr(1,path.size()-1);
      if(path.size()&&(path[path.size()-1]!='/')) path += "/";
      if(!fWriter.openEntry(path)) return false;
      return fWriter.closeEntry();
    }
    virtual bool endVisit(const Slash::Store::IFolder&,
                          const Slash::Store::IFolder& aFolder){ 
      if(fVerboseLevel) {
        Lib::Out out(fPrinter);
        out << "BatchLab::Zip_StoreVisitor::endVisit :" 
            << " path : " << Lib::sout(aFolder.path())
            << Lib::endl;
      }
      return true;
    }
    virtual bool isRecursive() const { return true;}
  private:
    Zip::Tree& fTree;  
    AIDA_ML fAIDA_ML;
    Lib::ZipWriter& fWriter;
    Slash::Core::IWriter& fPrinter;
    int fVerboseLevel;
};
}

//////////////////////////////////////////////////////////////////////////////
bool BatchLab::Zip::Tree::commit(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fReadOnly) return false;
  std::string fileName = name(); //!!!
  if(fCreateNew) {
    Lib::fmanip::remove(fileName);
  } else {
    //Can't update a file !
    return false;
  }

  // The name may contain env variables.
  std::string fullName = Lib::fmanip::name(fileName);
  Lib::ZipWriter* writer = new Lib::ZipWriter(fullName,printer());
  if(!writer->enabled()) {
    delete writer;
    return false;
  }
  
  Lib::ItemMLFactory factory; //FIXME
  Zip_StoreVisitor visitor(*this,factory,*writer,printer());
  rootFolder().visit(visitor);

  delete writer; //Close the file.

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::Zip::Tree::close(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::Zip::Tree::compress(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fCompress;
}

