#ifndef Lib_A4s_h
#define Lib_A4s_h

// Inheritance :
#include <Lib/Array.h>

namespace Lib {

// For 4 dimensional tensor manipulations (4-point, metric, Christoffel).
template <class T>
class A4 : public Array<T> {
public:
  A4():Array<T>(1,4){}
  A4(const A4& aFrom):Array<T>(aFrom){}
  virtual ~A4() {}
  inline void setValue(const T& a1,const T& a2,const T& a3,const T& a4) {
    Array<T>::fVector[0] = a1;
    Array<T>::fVector[1] = a2;
    Array<T>::fVector[2] = a3;
    Array<T>::fVector[3] = a4;
  }
  inline T value0() const { return Array<T>::fVector[0];}
  inline T value1() const { return Array<T>::fVector[1];}
  inline T value2() const { return Array<T>::fVector[2];}
  inline T value3() const { return Array<T>::fVector[3];}

  inline void setValue0(const T& aValue) { Array<T>::fVector[0] = aValue;}
  inline void setValue1(const T& aValue) { Array<T>::fVector[1] = aValue;}
  inline void setValue2(const T& aValue) { Array<T>::fVector[2] = aValue;}
  inline void setValue3(const T& aValue) { Array<T>::fVector[3] = aValue;}
};

#define A44_D33(a_11,a_12,a_13,a_21,a_22,a_23,a_31,a_32,a_33)\
 (a_11 * (a_22 * a_33 - a_32 * a_23)- a_21 * (a_12 * a_33 - a_32 * a_13) + a_31 * (a_12 * a_23 - a_22 * a_13))

template <class T>
class A44 : public ADD<T> {
public:
  A44():ADD<T>(4){}
  A44(const A44& aFrom):ADD<T>(aFrom){}
  virtual ~A44() {}
  inline A44& operator=(const A44& aFrom) {
    ADD<T>::operator=(aFrom);
    return *this;
  }

#define A44_SET_VALUE(aName,a1,a2) \
  inline void aName(const T& aValue) { Array<T>::fVector[a1 + a2 * 4] = aValue;}

  A44_SET_VALUE(setValue00,0,0)
  A44_SET_VALUE(setValue10,1,0)
  A44_SET_VALUE(setValue20,2,0)
  A44_SET_VALUE(setValue30,3,0)

  A44_SET_VALUE(setValue01,0,1)
  A44_SET_VALUE(setValue11,1,1)
  A44_SET_VALUE(setValue21,2,1)
  A44_SET_VALUE(setValue31,3,1)

  A44_SET_VALUE(setValue02,0,2)
  A44_SET_VALUE(setValue12,1,2)
  A44_SET_VALUE(setValue22,2,2)
  A44_SET_VALUE(setValue32,3,2)

  A44_SET_VALUE(setValue03,0,3)
  A44_SET_VALUE(setValue13,1,3)
  A44_SET_VALUE(setValue23,2,3)
  A44_SET_VALUE(setValue33,3,3)

#undef A44_SET_VALUE

#define A44_GET_VALUE(aName,a1,a2) \
  inline T aName() const { return Array<T>::fVector[a1 + a2 * 4];}

  A44_GET_VALUE(value00,0,0)
  A44_GET_VALUE(value10,1,0)
  A44_GET_VALUE(value20,2,0)
  A44_GET_VALUE(value30,3,0)

  A44_GET_VALUE(value01,0,1)
  A44_GET_VALUE(value11,1,1)
  A44_GET_VALUE(value21,2,1)
  A44_GET_VALUE(value31,3,1)

  A44_GET_VALUE(value02,0,2)
  A44_GET_VALUE(value12,1,2)
  A44_GET_VALUE(value22,2,2)
  A44_GET_VALUE(value32,3,2)

  A44_GET_VALUE(value03,0,3)
  A44_GET_VALUE(value13,1,3)
  A44_GET_VALUE(value23,2,3)
  A44_GET_VALUE(value33,3,3)

#undef A44_GET_VALUE

  inline virtual T determinant() const {
    T e00 =  A44_D33(value11(),value12(),value13(),
                      value21(),value22(),value23(),
                      value31(),value32(),value33());
    T e01 = -A44_D33(value01(),value02(),value03(),
                      value21(),value22(),value23(),
                      value31(),value32(),value33());
    T e02 =  A44_D33(value01(),value02(),value03(),
                      value11(),value12(),value13(),
                      value31(),value32(),value33());
    T e03 = -A44_D33(value01(),value02(),value03(),
                      value11(),value12(),value13(),
                      value21(),value22(),value23());
                  
    return (  value00() * e00
            + value10() * e01
            + value20() * e02
            + value30() * e03);
  }

  inline bool invert(A44<T>& aResult) const {
    // Twice faster than the generic invertion method.
    aResult.setValue00(
      A44_D33(value11(),value12(),value13(),
          value21(),value22(),value23(),
  	value31(),value32(),value33()));
    aResult.setValue01(
      -A44_D33(value01(),value02(),value03(),
  	 value21(),value22(),value23(),
           value31(),value32(),value33()));
    aResult.setValue02(
      A44_D33(value01(),value02(),value03(),
          value11(),value12(),value13(),
  	value31(),value32(),value33()));
    aResult.setValue03(
     -A44_D33(value01(),value02(),value03(),
          value11(),value12(),value13(),
  	value21(),value22(),value23()));
                    
    T deter =
        value00() * aResult.value00()
      + value10() * aResult.value01()
      + value20() * aResult.value02()
      + value30() * aResult.value03();
  
    if(deter==Array<T>::zero()) return false;
                    
    aResult.setValue10(
     -A44_D33(value10(),value12(),value13(),
          value20(),value22(),value23(),
          value30(),value32(),value33()));
    aResult.setValue11(
      A44_D33(value00(),value02(),value03(),
  	value20(),value22(),value23(),
  	value30(),value32(),value33()));
    aResult.setValue12(
     -A44_D33(value00(),value02(),value03(),
          value10(),value12(),value13(),
  	value30(),value32(),value33()));
    aResult.setValue13(
      A44_D33(value00(),value02(),value03(),
  	value10(),value12(),value13(),
  	value20(),value22(),value23()));
                    
    aResult.setValue20(
      A44_D33(value10(),value11(),value13(),
  	value20(),value21(),value23(),
  	value30(),value31(),value33()));
    aResult.setValue21(
     -A44_D33(value00(),value01(),value03(),
  	value20(),value21(),value23(),
  	value30(),value31(),value33()));
    aResult.setValue22(
      A44_D33(value00(),value01(),value03(),
  	value10(),value11(),value13(),
  	value30(),value31(),value33()));
    aResult.setValue23(
     -A44_D33(value00(),value01(),value03(),
  	value10(),value11(),value13(),
  	value20(),value21(),value23()));
                    
    aResult.setValue30(
     -A44_D33(value10(),value11(),value12(),
  	value20(),value21(),value22(),
  	value30(),value31(),value32()));
    aResult.setValue31(
      A44_D33(value00(),value01(),value02(),
  	value20(),value21(),value22(),
  	value30(),value31(),value32()));
    aResult.setValue32(
     -A44_D33(value00(),value01(),value02(),
  	value10(),value11(),value12(),
  	value30(),value31(),value32()));
    aResult.setValue33(
      A44_D33(value00(),value01(),value02(),
  	value10(),value11(),value12(),
  	value20(),value21(),value22()));
  
    aResult.multiply(1./deter);  
  
    return true;
  }
#undef A44_D33

  
};

template <class T>
class A444 : public Array<T> {
public:
  A444():Array<T>(3,4){}
  A444(const A444& aFrom):Array<T>(aFrom){}
  virtual ~A444() {}

  inline T value(unsigned int a1,unsigned int a2,unsigned int a3) const { 
    return Array<T>::fVector[a1 + a2 * 4 + a3 * 16];
  }

  inline void setValue(unsigned int a1,unsigned int a2,unsigned int a3,const T& aValue) { 
    Array<T>::fVector[a1 + a2 * 4 + a3 * 16] = aValue;
  }

#define A444_SET_VALUE(aName,a1,a2,a3) \
  inline void aName(const T& aValue) { Array<T>::fVector[a1 + a2 * 4 + a3 * 16] = aValue;}

  A444_SET_VALUE(setValue000,0,0,0)
  A444_SET_VALUE(setValue100,1,0,0)
  A444_SET_VALUE(setValue200,2,0,0)
  A444_SET_VALUE(setValue300,3,0,0)
  A444_SET_VALUE(setValue010,0,1,0)
  A444_SET_VALUE(setValue110,1,1,0)
  A444_SET_VALUE(setValue210,2,1,0)
  A444_SET_VALUE(setValue310,3,1,0)
  A444_SET_VALUE(setValue020,0,2,0)
  A444_SET_VALUE(setValue120,1,2,0)
  A444_SET_VALUE(setValue220,2,2,0)
  A444_SET_VALUE(setValue320,3,2,0)
  A444_SET_VALUE(setValue030,0,3,0)
  A444_SET_VALUE(setValue130,1,3,0)
  A444_SET_VALUE(setValue230,2,3,0)
  A444_SET_VALUE(setValue330,3,3,0)

  A444_SET_VALUE(setValue001,0,0,1)
  A444_SET_VALUE(setValue101,1,0,1)
  A444_SET_VALUE(setValue201,2,0,1)
  A444_SET_VALUE(setValue301,3,0,1)
  A444_SET_VALUE(setValue011,0,1,1)
  A444_SET_VALUE(setValue111,1,1,1)
  A444_SET_VALUE(setValue211,2,1,1)
  A444_SET_VALUE(setValue311,3,1,1)
  A444_SET_VALUE(setValue021,0,2,1)
  A444_SET_VALUE(setValue121,1,2,1)
  A444_SET_VALUE(setValue221,2,2,1)
  A444_SET_VALUE(setValue321,3,2,1)
  A444_SET_VALUE(setValue031,0,3,1)
  A444_SET_VALUE(setValue131,1,3,1)
  A444_SET_VALUE(setValue231,2,3,1)
  A444_SET_VALUE(setValue331,3,3,1)

  A444_SET_VALUE(setValue002,0,0,2)
  A444_SET_VALUE(setValue102,1,0,2)
  A444_SET_VALUE(setValue202,2,0,2)
  A444_SET_VALUE(setValue302,3,0,2)
  A444_SET_VALUE(setValue012,0,1,2)
  A444_SET_VALUE(setValue112,1,1,2)
  A444_SET_VALUE(setValue212,2,1,2)
  A444_SET_VALUE(setValue312,3,1,2)
  A444_SET_VALUE(setValue022,0,2,2)
  A444_SET_VALUE(setValue122,1,2,2)
  A444_SET_VALUE(setValue222,2,2,2)
  A444_SET_VALUE(setValue322,3,2,2)
  A444_SET_VALUE(setValue032,0,3,2)
  A444_SET_VALUE(setValue132,1,3,2)
  A444_SET_VALUE(setValue232,2,3,2)
  A444_SET_VALUE(setValue332,3,3,2)

  A444_SET_VALUE(setValue003,0,0,3)
  A444_SET_VALUE(setValue103,1,0,3)
  A444_SET_VALUE(setValue203,2,0,3)
  A444_SET_VALUE(setValue303,3,0,3)
  A444_SET_VALUE(setValue013,0,1,3)
  A444_SET_VALUE(setValue113,1,1,3)
  A444_SET_VALUE(setValue213,2,1,3)
  A444_SET_VALUE(setValue313,3,1,3)
  A444_SET_VALUE(setValue023,0,2,3)
  A444_SET_VALUE(setValue123,1,2,3)
  A444_SET_VALUE(setValue223,2,2,3)
  A444_SET_VALUE(setValue323,3,2,3)
  A444_SET_VALUE(setValue033,0,3,3)
  A444_SET_VALUE(setValue133,1,3,3)
  A444_SET_VALUE(setValue233,2,3,3)
  A444_SET_VALUE(setValue333,3,3,3)

#undef A444_SET_VALUE

};

}

template <class T>
Lib::A44<T> operator-(const Lib::A44<T>& a1,const Lib::A44<T>& a2) {
  Lib::A44<T> res(a1);
  res.subtract(a2);
  return res;
}
template <class T>
Lib::A44<T> operator+(const Lib::A44<T>& a1,const Lib::A44<T>& a2) {
  Lib::A44<T> res(a1);
  res.add(a2);
  return res;
}
template <class T>
Lib::A44<T> operator*(const Lib::A44<T>& a1,const Lib::A44<T>& a2) {
  Lib::A44<T> res;
  a1.mxmul(a2,res);
  return res;
}

namespace Lib {

template <class T>
A44<T> commutator(const A44<T>& a1,const A44<T>& a2) {
  return a1*a2-a2*a1;
}
template <class T>
A44<T> antiCommutator(const A44<T>& a1,const A44<T>& a2) {
  return a1*a2+a2*a1;
}


//////////////////////////////////////////////////////////////////////////////
/// Some common real matrices ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

typedef A4<double> A4d;
typedef A44<double> A44d;
typedef A444<double> A444d;

bool l2(const A4d&,const A44d&,double&);
bool x0normalize(A4d&,const A44d&);

//  1  0  0  0
//  0  1  0  0
//  0  0  1  0
//  0  0  0  1
class A44d_I : public A44d {
public:
 A44d_I();
};

//  1  0  0  0
//  0 -1  0  0
//  0  0 -1  0
//  0  0  0 -1
class A44d_eta : public A44d {
public:
 A44d_eta();
};

//  0  0  0  1
//  0  0 -1  0
//  0  1  0  0
// -1  0  0  0
class A44d_J : public A44d {
public:
 A44d_J();
};

class A44d_K : public A44d {
public:
 A44d_K();
};
class A44d_T1 : public A44d {
public:
 A44d_T1();
};
class A44d_T2 : public A44d {
public:
 A44d_T2();
};
class A44d_T3 : public A44d {
public:
 A44d_T3();
};

//////////////////////////////////////////////////////////////////////////////
/// Some common complex matrices /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
typedef A44< std::complex<double> > A44z;

class A44z_I : public A44z {
public:
 A44z_I();
};
class A44z_Dirac0 : public A44z {
public:
 A44z_Dirac0();
};
class A44z_Dirac1 : public A44z {
public:
 A44z_Dirac1();
};
class A44z_Dirac2 : public A44z {
public:
 A44z_Dirac2();
};
class A44z_Dirac3 : public A44z {
public:
 A44z_Dirac3();
};

}

#endif




