
#include <Lib/Printer.h>
#include <Lib/Out.h>
#include <Lib/smanip.h>

#include <Lib/As.h>
#include <Lib/A2s.h>
#include <Lib/A3s.h>
#include <Lib/A4s.h>

typedef std::complex<double> z;

double depsilon = 1e-10;

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void check_paulis(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A22z_Pauli1 pauli1;
  Lib::A22z_Pauli2 pauli2;
  Lib::A22z_Pauli3 pauli3;

  checkInvert(pauli1,printer);
  checkInvert(pauli2,printer);
  checkInvert(pauli3,printer);
}
//////////////////////////////////////////////////////////////////////////////
void check_eta_J(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_J J;
  checkInvert(J,printer);

  Lib::A44d_eta eta;
  checkInvert(eta,printer);

  Lib::A44d tmp = J + eta;
 
  Lib::ADD<double> stmp = tmp.symmetricPart();
  if(!stmp.equal(eta)) {
    out << "Problem with sym part of J+eta." << Lib::endl;
  }
  
  Lib::ADD<double> atmp = tmp.antiSymmetricPart();
  if(!atmp.equal(J)) {
    out << "Problem with antisym part of J+eta." << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void check_paulis_commutators(
) 
//////////////////////////////////////////////////////////////////////////////
// [pauli(j),pauli(k)] = 2 * i epsilon(j,k,l) * pauli(l)
// and then :
// [i*pauli(j)/2,i*pauli(k)/2] = - epsilon(j,k,l) * (i*pauli(l)/2)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A22z_Pauli1 pauli1;
  Lib::A22z_Pauli2 pauli2;
  Lib::A22z_Pauli3 pauli3;

  std::vector<Lib::A22z> paulis;
  paulis.push_back(pauli1);
  paulis.push_back(pauli2);
  paulis.push_back(pauli3);

  unsigned int paulin = paulis.size();
  Lib::A333d_epsilon epsilon;
  z zi(0,1);
  for(unsigned int j=0;j<paulin;j++) {
    for(unsigned int k=j;k<paulin;k++) {
      Lib::A22z cm = Lib::commutator(paulis[j],paulis[k]);
      Lib::A22z tmp;
      for(unsigned int l=0;l<paulin;l++) {
        Lib::A22z tmp1(paulis[l]);
        tmp1.multiply(epsilon.value(j,k,l));
        tmp += tmp1;
      }
      tmp.multiply(zi);
      tmp.multiply(2);
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with [pauli%d,pauli%d]",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void check_paulis_anticommutators(
) 
//////////////////////////////////////////////////////////////////////////////
// {pauli(j),pauli(k)} = 2 delta(j,k) * id
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A22z_Pauli1 pauli1;
  Lib::A22z_Pauli2 pauli2;
  Lib::A22z_Pauli3 pauli3;

  std::vector<Lib::A22z> paulis;
  paulis.push_back(pauli1);
  paulis.push_back(pauli2);
  paulis.push_back(pauli3);

  unsigned int paulin = paulis.size();

  Lib::A22z_I id;
  Lib::A33d_I delta;
  for(unsigned int j=0;j<paulin;j++) {
    for(unsigned int k=j;k<paulin;k++) {
      Lib::A22z ac = Lib::antiCommutator(paulis[j],paulis[k]);
      Lib::A22z tmp(id);
      tmp.multiply(z(2*delta.value(j,k),0));
      if(!ac.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with {pauli%d,pauli%d}",j,k);
        Lib::print(ac,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }

}
//////////////////////////////////////////////////////////////////////////////
static Lib::A22z weyl_commutator(
 const Lib::A22z& a1
,const Lib::A22z& a2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::A22z_Epsilon eps;

  Lib::A22z tmp1 = a1;
  tmp1 *= eps;
 {Lib::A22z tmp = a2;
  Lib::conj(tmp);
  tmp1 *= tmp;}

  Lib::A22z tmp2 = a2;
  tmp2 *= eps;
 {Lib::A22z tmp = a1;
  Lib::conj(tmp);
  tmp2 *= tmp;}

  Lib::A22z tmp = tmp1-tmp2;
  tmp *= eps;

  return tmp;
}
//////////////////////////////////////////////////////////////////////////////
static Lib::A22z weyl_antiCommutator(
 const Lib::A22z& a1
,const Lib::A22z& a2
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::A22z_Epsilon eps;

  Lib::A22z tmp1 = a1;
  tmp1 *= eps;
 {Lib::A22z tmp = a2;
  Lib::conj(tmp);
  tmp1 *= tmp;}

  Lib::A22z tmp2 = a2;
  tmp2 *= eps;
 {Lib::A22z tmp = a1;
  Lib::conj(tmp);
  tmp2 *= tmp;}

  Lib::A22z tmp = tmp1+tmp2;
  tmp *= eps;

  return tmp;
}
//////////////////////////////////////////////////////////////////////////////
void check_weyl_generators(
)
//////////////////////////////////////////////////////////////////////////////
// weyl_alpha = (identity,pauli_1,pauli_2,pauli_3)
// gen_alpha_beta = ( weyl_alpha*J*weyl_beta_conj*J - 
//                    weyl_beta*J*weyl_alpha_conj*J )
// We must have :
// [gen_mu_nu,gen_alpha_beta] = -4 * (
//   -eta(mu,alpha)*gen_nu_beta + eta(mu,beta)*gen_nu_alpha
//   -eta(nu,beta)*gen_mu_alpha + eta(nu,alpha)*gen_mu_beta )
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A22z> weyls;
  weyls.push_back(Lib::A22z_I());
  weyls.push_back(Lib::A22z_Pauli1());
  weyls.push_back(Lib::A22z_Pauli2());
  weyls.push_back(Lib::A22z_Pauli3());

  //weyls[1].multiply(-1);
  //weyls[2].multiply(-1);
  //weyls[3].multiply(-1);

  Lib::A44d_eta eta;

 {for(unsigned int mu=0;mu<4;mu++) {
   for(unsigned int nu=0;nu<4;nu++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=0;beta<4;beta++) {
        Lib::A22z cmt1 = weyl_commutator(weyls[mu],weyls[nu]);
        //cmt1.multiply(1./8.);
        Lib::A22z cmt2 = weyl_commutator(weyls[alpha],weyls[beta]);
        //cmt2.multiply(1./8.);
        Lib::A22z cmt = Lib::commutator(cmt1,cmt2);

        Lib::A22z tmp;
       {Lib::A22z tmp1 = weyl_commutator(weyls[nu],weyls[beta]);
        tmp1.multiply(-eta.value(mu,alpha));
        Lib::A22z tmp2 = weyl_commutator(weyls[nu],weyls[alpha]);
        tmp2.multiply( eta.value(mu,beta));
        tmp = tmp1 + tmp2;}

       {Lib::A22z tmp1 = weyl_commutator(weyls[mu],weyls[alpha]);
        tmp1.multiply(-eta.value(nu,beta));
        Lib::A22z tmp2 = weyl_commutator(weyls[mu],weyls[beta]);
        tmp2.multiply( eta.value(nu,alpha));
        tmp = tmp + tmp1 + tmp2;}

        tmp.multiply(-4);

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
            (scmt,128," problem check_weyl_generators [%d_%d,%d_%d] :",
               mu,nu,alpha,beta);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
   }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_weyl_generators_anti(
)
//////////////////////////////////////////////////////////////////////////////
// weyl_alpha = (identity,pauli_1,pauli_2,pauli_3)
// gen_anti_alpha_beta = ( weyl_alpha*J*weyl_beta_conj*J + 
//                         weyl_beta*J*weyl_alpha_conj*J )
// We must have :
//  gen_anti_alpha_beta = -2*eta(alpha,beta)*I
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A22z> weyls;
  weyls.push_back(Lib::A22z_I());
  weyls.push_back(Lib::A22z_Pauli1());
  weyls.push_back(Lib::A22z_Pauli2());
  weyls.push_back(Lib::A22z_Pauli3());

  //weyls[1].multiply(-1);
  //weyls[2].multiply(-1);
  //weyls[3].multiply(-1);

  Lib::A44d_eta eta;

  Lib::A22z_I id;

 {for(unsigned int alpha=0;alpha<4;alpha++) {
    for(unsigned int beta=0;beta<4;beta++) {
      Lib::A22z cmt = weyl_antiCommutator(weyls[alpha],weyls[beta]);

      Lib::A22z tmp = id;
      tmp.multiply(-2.*eta.value(alpha,beta));

      if(!tmp.equal(cmt)) {
        std:: string scmt;
        Lib::smanip::printf
          (scmt,128," problem check_weyl_generators_anti %d %d :",
             alpha,beta);
        Lib::print(cmt,printer,scmt);
        Lib::print(tmp,printer,"");
      }
   }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_weyl(
)
//////////////////////////////////////////////////////////////////////////////
// weyl_alpha = (identity,pauli_1,pauli_2,pauli_3)
// We have : with i,j,k = 1,2,3 (not zero).
//   [[weyl_i,weyl_j],weyl_k] = 4*[ eta(i,k)*weyl_j - eta(j,k)*weyl_i ] 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A22z> weyls;
  weyls.push_back(Lib::A22z_I());
  weyls.push_back(Lib::A22z_Pauli1());
  weyls.push_back(Lib::A22z_Pauli2());
  weyls.push_back(Lib::A22z_Pauli3());

  weyls[1].multiply(-1);
  weyls[2].multiply(-1);
  weyls[3].multiply(-1);

  Lib::A44d_eta eta;

 {for(unsigned int gamma=1;gamma<4;gamma++) {
    for(unsigned int alpha=1;alpha<4;alpha++) {
      for(unsigned int beta=1;beta<4;beta++) {
        Lib::A22z cmt1 = Lib::commutator(weyls[alpha],weyls[beta]);

        Lib::A22z cmt = Lib::commutator(cmt1,weyls[gamma]);

        Lib::A22z tmp;
       {Lib::A22z tmp1 = weyls[beta];
        tmp1.multiply( 4.*eta.value(alpha,gamma));
        Lib::A22z tmp2 = weyls[alpha];
        tmp2.multiply(-4.*eta.value(beta,gamma));
        tmp = tmp1 + tmp2;}

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
            (scmt,128," problem check_weyl [[%d,%d],%d] :",alpha,beta,gamma);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_weyl_SUSY(
)
//////////////////////////////////////////////////////////////////////////////
// weyl_alpha = (identity,pauli_1,pauli_2,pauli_3)
// weyl_com_alpha_beta = weyl_commutator(alpha,beta)
// We have : 
//  weyl_com_alpha_beta * weyl_gamma + weyl_gamma * weyl_com_alpha_beta.dagger
//  = 4*[ eta(alpha,gamma)*weyl_beta - eta(beta,gamma)*weyl_alpha ] 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A22z> weyls;
  weyls.push_back(Lib::A22z_I());
  weyls.push_back(Lib::A22z_Pauli1());
  weyls.push_back(Lib::A22z_Pauli2());
  weyls.push_back(Lib::A22z_Pauli3());

  //weyls[1].multiply(-1);
  //weyls[2].multiply(-1);
  //weyls[3].multiply(-1);

  Lib::A44d_eta eta;

 {for(unsigned int gamma=0;gamma<4;gamma++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=alpha;beta<4;beta++) {
        Lib::A22z cmt1 = weyl_commutator(weyls[alpha],weyls[beta]);

        Lib::A22z tmp1 = cmt1;
        tmp1 *= weyls[gamma];
        Lib::A22z tmp2 = weyls[gamma];
        Lib::A22z tmp3 = cmt1;
        Lib::dagger(tmp3);
        tmp2 *= tmp3;
        Lib::A22z cmt = tmp1 + tmp2;

        Lib::A22z tmp;
       {Lib::A22z tmp1 = weyls[beta];
        tmp1.multiply( 4.*eta.value(alpha,gamma));
        Lib::A22z tmp2 = weyls[alpha];
        tmp2.multiply(-4.*eta.value(beta,gamma));
        tmp = tmp1 + tmp2;}

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
            (scmt,128," problem check_weyl_SUSY (%d,%d) %d :", 
              alpha,beta,gamma);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_weyl_SUSY_2(
)
//////////////////////////////////////////////////////////////////////////////
// weyl_alpha = (identity,pauli_1,pauli_2,pauli_3)
// weyl_com_alpha_beta = weyl_commutator(alpha,beta)
// We have : 
//  weyl_com_alpha_beta.dagger * weyl_gamma + weyl_gamma * weyl_com_alpha_beta
//  = -4*[ delta(alpha,gamma)*eta(beta,u)*weyl_u - 
//         delta(beta,gamma)*eta(alpha,u)*weyl_u ] 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A22z> weyls;
  weyls.push_back(Lib::A22z_I());
  weyls.push_back(Lib::A22z_Pauli1());
  weyls.push_back(Lib::A22z_Pauli2());
  weyls.push_back(Lib::A22z_Pauli3());

  //weyls[1].multiply(-1);
  //weyls[2].multiply(-1);
  //weyls[3].multiply(-1);

  Lib::A44d_eta eta;
  Lib::A44d_I id;

 {for(unsigned int gamma=0;gamma<4;gamma++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=alpha;beta<4;beta++) {
        Lib::A22z cmt1 = weyl_commutator(weyls[alpha],weyls[beta]);

        Lib::A22z tmp1 = cmt1;
        Lib::dagger(tmp1);
        tmp1 *= weyls[gamma];
        Lib::A22z tmp2 = weyls[gamma];
        tmp2 *= cmt1;
        Lib::A22z cmt = tmp1 + tmp2;

        Lib::A22z tmp;
       {for(unsigned int u=0;u<4;u++) {
          Lib::A22z tmp1 = weyls[u];
          tmp1.multiply( 4.*id.value(alpha,gamma)*eta.value(beta,u));
          tmp += tmp1;
        }}
       {for(unsigned int u=0;u<4;u++) {
          Lib::A22z tmp1 = weyls[u];
          tmp1.multiply(-4.*id.value(beta,gamma)*eta.value(alpha,u));
          tmp += tmp1;
        }}

        tmp.multiply(-1);

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
            (scmt,128," problem check_weyl_SUSY_2 (%d,%d) %d :", 
              alpha,beta,gamma);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_dirac_SUSY_dirac(
)
//////////////////////////////////////////////////////////////////////////////
// dirac_com_alpha_beta = eta(alpha,mu)*eta(beta,nu)*[dirac_mu,dirac_nu]
// We have : 
//  dirac_com_alpha_beta * dirac_gamma - dirac_gamma * dirac_com_alpha_beta
//  = -4*[ delta(alpha,gamma)*eta(beta,u)*dirac_u - 
//         delta(beta,gamma)*eta(alpha,u)*dirac_u ] 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  Lib::A44d_eta eta;
  Lib::A44d_I id;

 {for(unsigned int gamma=0;gamma<4;gamma++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=alpha;beta<4;beta++) {
        Lib::A44z cmt1;
       {for(unsigned int u=0;u<4;u++) {
          for(unsigned int v=0;v<4;v++) {
            Lib::A44z tmp = Lib::commutator(diracs[u],diracs[v]);
            tmp.multiply(eta.value(alpha,u));
            tmp.multiply(eta.value(beta,v));
            cmt1 += tmp;
          }
        }}

        Lib::A44z cmt = Lib::commutator(cmt1,diracs[gamma]);

        Lib::A44z tmp;
       {for(unsigned int u=0;u<4;u++) {
          Lib::A44z tmp1 = diracs[u];
          tmp1.multiply( 4.*id.value(alpha,gamma)*eta.value(beta,u));
          tmp += tmp1;
        }}
       {for(unsigned int u=0;u<4;u++) {
          Lib::A44z tmp1 = diracs[u];
          tmp1.multiply(-4.*id.value(beta,gamma)*eta.value(alpha,u));
          tmp += tmp1;
        }}

        tmp.multiply(-1);

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
            (scmt,128," problem check_dirac_SUSY_dirac (%d,%d) %d :", 
              alpha,beta,gamma);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_dirac_lorentz_SUSY_generators(
)
//////////////////////////////////////////////////////////////////////////////
// gen_alpha_beta = eta(alpha,mu)*eta(beta,nu)*[dirac_mu,dirac_nu]
// We have NOT :
// [gen_mu_nu,gen_alpha_beta] = -4 * (
//   -eta(mu,alpha)*gen_nu_beta + eta(mu,beta)*gen_nu_alpha
//   -eta(nu,beta)*gen_mu_alpha + eta(nu,alpha)*gen_mu_beta )
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  Lib::A44d_eta eta;

 {for(unsigned int mu=0;mu<4;mu++) {
   for(unsigned int nu=0;nu<4;nu++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=0;beta<4;beta++) {

        Lib::A44z cmt1;
       {for(unsigned int u=0;u<4;u++) {
          for(unsigned int v=0;v<4;v++) {
            Lib::A44z tmp = Lib::commutator(diracs[u],diracs[v]);
            tmp.multiply(eta.value(mu,u));
            tmp.multiply(eta.value(nu,v));
            cmt1 += tmp;
          }
        }}

        Lib::A44z cmt2;
       {for(unsigned int u=0;u<4;u++) {
          for(unsigned int v=0;v<4;v++) {
            Lib::A44z tmp = Lib::commutator(diracs[u],diracs[v]);
            tmp.multiply(eta.value(alpha,u));
            tmp.multiply(eta.value(beta,v));
            cmt2 += tmp;
          }
        }}

        Lib::A44z cmt = Lib::commutator(cmt1,cmt2);

        Lib::A44z tmp;
       {Lib::A44z tmp1 = Lib::commutator(diracs[nu],diracs[beta]);
        tmp1.multiply(-eta.value(mu,alpha));
        Lib::A44z tmp2 = Lib::commutator(diracs[nu],diracs[alpha]);
        tmp2.multiply( eta.value(mu,beta));
        tmp = tmp1 + tmp2;}

       {Lib::A44z tmp1 = Lib::commutator(diracs[mu],diracs[alpha]);
        tmp1.multiply(-eta.value(nu,beta));
        Lib::A44z tmp2 = Lib::commutator(diracs[mu],diracs[beta]);
        tmp2.multiply( eta.value(nu,alpha));
        tmp = tmp + tmp1 + tmp2;}

        tmp.multiply(4);

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
           (scmt,128," problem check_dirac_lorentz_SUSY_generators [%d_%d,%d_%d] :",
               mu,nu,alpha,beta);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
   }
  }}

}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void check_taus_anticommutators(
) 
//////////////////////////////////////////////////////////////////////////////
//   {tau(alpha),tau(beta)} = -2 * eta(alpha,beta) * id
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_J J;
  Lib::A44d_eta eta;

  Lib::A44d_K K;
  Lib::A44d_I id;
  std::vector<Lib::A44d> taus;
  Lib::A44d T0 = K + J * K;
  T0.multiply(1./::sqrt(2.));
  taus.push_back(T0);
  taus.push_back(Lib::A44d_T1());
  taus.push_back(Lib::A44d_T2());
  taus.push_back(Lib::A44d_T3());
  unsigned int taun = taus.size();
  for(unsigned int j=0;j<taun;j++) {
    for(unsigned int k=j;k<taun;k++) {
      Lib::A44d ac = Lib::antiCommutator(taus[j],taus[k]);
      Lib::A44d tmp(id);
      tmp.multiply(-2*eta.value(j,k));
      if(!ac.equal(tmp,depsilon)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with {tau%d,tau%d}",j,k);
        Lib::print(ac,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }

}
//////////////////////////////////////////////////////////////////////////////
void check_taus_commutators(
) 
//////////////////////////////////////////////////////////////////////////////
// On T1,T2,T3 only :
//   [tau(j),tau(k)] = 2 * J * epsilon(j,k,l) * tau(l)
// and then :
//   [J*tau(j)/2,J*tau(k)/2] = - epsilon(j,k,l) * (J*tau(l)/2)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_J J;

  Lib::A44d_T1 T1;
  Lib::A44d_T2 T2;
  Lib::A44d_T3 T3;

  std::vector<Lib::A44d> taus;
  taus.push_back(T1);
  taus.push_back(T2);
  taus.push_back(T3);

  unsigned int taun = taus.size();
  Lib::A333d_epsilon epsilon;
  for(unsigned int j=0;j<taun;j++) {
    for(unsigned int k=j;k<taun;k++) {
      Lib::A44d cm = Lib::commutator(taus[j],taus[k]);
      Lib::A44d tmp;
      for(unsigned int l=0;l<taun;l++) {
        Lib::A44d tmp1 = J * taus[l];
        tmp1.multiply(epsilon.value(j,k,l));
        tmp += tmp1;
      }
      tmp.multiply(2);
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with [tau%d,tau%d]",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void check_taus_transpose(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_J J;
  if(!J.isAntiSymmetric()) {
    out << "found J not antisymmetric." << Lib::endl;
  }
  Lib::A44d tmp(J);
  tmp.transpose();
  tmp.multiply(-1);
  if(!tmp.equal(J)) {
    out << "Trans(J) not -J." << Lib::endl;
  }

  Lib::A44d_T1 T1;
  Lib::A44d_T2 T2;
  Lib::A44d_T3 T3;

  std::vector<Lib::A44d> taus;
  taus.push_back(T1);
  taus.push_back(T2);
  taus.push_back(T3);

  unsigned int taun = taus.size();
  for(unsigned int j=0;j<taun;j++) {
    if(!taus[j].isSymmetric()) {
      out << "found tau" << j << " not symmetric." << Lib::endl;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void check_taus_J_commutators(
) 
//////////////////////////////////////////////////////////////////////////////
// On T1,T2,T3 only :
//   [tau(j),J] = 0
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_J J;

  Lib::A44d_T1 T1;
  Lib::A44d_T2 T2;
  Lib::A44d_T3 T3;

  std::vector<Lib::A44d> taus;
  taus.push_back(T1);
  taus.push_back(T2);
  taus.push_back(T3);

  unsigned int taun = taus.size();
  for(unsigned int j=0;j<taun;j++) {
    for(unsigned int k=j;k<taun;k++) {
      Lib::A44d cm = Lib::commutator(taus[j],J);
      Lib::A44d zero;
      if(!cm.equal(zero)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with [tau%d,J]",j);
        Lib::print(cm,printer,cmt);
      }
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void check_diracs_anticommutators(
) 
//////////////////////////////////////////////////////////////////////////////
//   {dirac(alpha),dirac(beta)} = 2 * eta(alpha,beta) * Id
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_eta eta;

  Lib::A44z_I id;
  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  unsigned int diracn = diracs.size();
  for(unsigned int j=0;j<diracn;j++) {
    for(unsigned int k=j;k<diracn;k++) {
      Lib::A44z ac = Lib::antiCommutator(diracs[j],diracs[k]);
      Lib::A44z tmp(id);
      tmp.multiply(z(2*eta.value(j,k),0));
      if(!ac.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with {dirac%d,dirac%d}",j,k);
        Lib::print(ac,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
bool check_dirac_commutator_commutator(
) 
//////////////////////////////////////////////////////////////////////////////
// Spinor representation of the GB group :
//   [[dirac(alpha),dirac(beta)],dirac(mu)] = 
//    4*(  -eta(alpha,mu)*dirac(beta) + eta(beta,mu)*dirac(alpha) )
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 4;
  unsigned int number = order * (order-1)/2;

  std::vector<Lib::A44z> reps;

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  Lib::A44d_eta eta;

  for(unsigned int alpha=0;alpha<4;alpha++) {
    for(unsigned int beta=0;beta<4;beta++) {
      for(unsigned int mu=0;mu<4;mu++) {
        Lib::A44z cm = Lib::commutator(diracs[alpha],diracs[beta]);
        Lib::A44z cm2 = Lib::commutator(cm,diracs[mu]);

        Lib::A44z tmp(diracs[beta]);
        tmp.multiply(z(-1*eta.value(alpha,mu),0));
        Lib::A44z tmp2(diracs[alpha]);
        tmp2.multiply(z(eta.value(beta,mu),0));

        tmp += tmp2;
        tmp.multiply(4);

        if(!cm2.equal(tmp)) {
          std:: string cmt;
          Lib::smanip::printf
          (cmt,128," problem with [[dirac%d,dirac%d],dirac%d]",alpha,beta,mu);
          Lib::print(cm2,printer,cmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
void check_vector_lorentz_generators(
)
//////////////////////////////////////////////////////////////////////////////
// (gen_mu_nu)alpha_beta = 
//   delta(alpha,mu)*eta(beta,nu)-delta(alpa,nu)*eta(beta,mu)
// We must have :
// [gen_mu_nu,gen_alpha_beta] =
//   -eta(mu,alpha)*gen_nu_beta + eta(mu,beta)*gen_nu_alpha
//   -eta(nu,beta)*gen_mu_alpha + eta(nu,alpha)*gen_mu_beta
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector< std::vector<Lib::A44d> > gens;

  Lib::A44d_eta eta;
  Lib::A44d_I id;

  gens.resize(4);
 {for(unsigned int mu=0;mu<4;mu++) {
    gens[mu].resize(4);
    for(unsigned int nu=0;nu<4;nu++) {
      Lib::A44d& tmp = gens[mu][nu];
      for(unsigned int alpha=0;alpha<4;alpha++) {
        for(unsigned int beta=0;beta<4;beta++) {
          double value = id.value(alpha,mu)*eta.value(beta,nu)-
                         id.value(alpha,nu)*eta.value(beta,mu);
          tmp.setValue(alpha,beta,value);
        }
      }
    }
  }}

 {for(unsigned int mu=0;mu<4;mu++) {
    for(unsigned int nu=0;nu<4;nu++) {
      for(unsigned int alpha=0;alpha<4;alpha++) {
        for(unsigned int beta=0;beta<4;beta++) {
          Lib::A44d gen1 = gens[mu][nu];
          Lib::A44d gen2 = gens[alpha][beta];
          Lib::A44d cmt = Lib::commutator(gen1,gen2);

          Lib::A44d tmp;
         {Lib::A44d tmp1 = gens[nu][beta];
          tmp1.multiply(-eta.value(mu,alpha));
          Lib::A44d tmp2 = gens[nu][alpha];
          tmp2.multiply( eta.value(mu,beta));
          tmp = tmp1 + tmp2;}

         {Lib::A44d tmp1 = gens[mu][alpha];
          tmp1.multiply(-eta.value(nu,beta));
          Lib::A44d tmp2 = gens[mu][beta];
          tmp2.multiply( eta.value(nu,alpha));
          tmp = tmp + tmp1 + tmp2;}

          //tmp.multiply(-1);

          if(!tmp.equal(cmt)) {
            std:: string scmt;
            Lib::smanip::printf
             (scmt,128,
               " problem check_vector_lorentz_generators [%d_%d,%d_%d] :",
               mu,nu,alpha,beta);
            Lib::print(cmt,printer,scmt);
            Lib::print(tmp,printer,"");
          }

        }
      }
    }
  }}


}
//////////////////////////////////////////////////////////////////////////////
void check_dirac_lorentz_generators(
)
//////////////////////////////////////////////////////////////////////////////
// gen_alpha_beta = [dirac_alpha,dirac_beta]
// We must have :
// [gen_mu_nu,gen_alpha_beta] = 4 * (
//   -eta(mu,alpha)*gen_nu_beta + eta(mu,beta)*gen_nu_alpha
//   -eta(nu,beta)*gen_mu_alpha + eta(nu,alpha)*gen_mu_beta )
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  Lib::A44d_eta eta;

 {for(unsigned int mu=0;mu<4;mu++) {
   for(unsigned int nu=0;nu<4;nu++) {
    for(unsigned int alpha=0;alpha<4;alpha++) {
      for(unsigned int beta=0;beta<4;beta++) {
        Lib::A44z cmt1 = Lib::commutator(diracs[mu],diracs[nu]);
        Lib::A44z cmt2 = Lib::commutator(diracs[alpha],diracs[beta]);
        Lib::A44z cmt = Lib::commutator(cmt1,cmt2);

        Lib::A44z tmp;
       {Lib::A44z tmp1 = Lib::commutator(diracs[nu],diracs[beta]);
        tmp1.multiply(-eta.value(mu,alpha));
        Lib::A44z tmp2 = Lib::commutator(diracs[nu],diracs[alpha]);
        tmp2.multiply( eta.value(mu,beta));
        tmp = tmp1 + tmp2;}

       {Lib::A44z tmp1 = Lib::commutator(diracs[mu],diracs[alpha]);
        tmp1.multiply(-eta.value(nu,beta));
        Lib::A44z tmp2 = Lib::commutator(diracs[mu],diracs[beta]);
        tmp2.multiply( eta.value(nu,alpha));
        tmp = tmp + tmp1 + tmp2;}

        tmp.multiply(4);

        if(!tmp.equal(cmt)) {
          std:: string scmt;
          Lib::smanip::printf
           (scmt,128," problem check_dirac_lorentz_generators [%d_%d,%d_%d] :",
               mu,nu,alpha,beta);
          Lib::print(cmt,printer,scmt);
          Lib::print(tmp,printer,"");
        }

      }
    }
   }
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_dirac_dagger(
) 
//////////////////////////////////////////////////////////////////////////////
//   dirac_0+ = dirac_0
//   dirac_k+ = -dirac_k
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

 {std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());
  unsigned int diracn = diracs.size();
  for(unsigned int j=0;j<diracn;j++) {
    Lib::A44z tmp = diracs[j];
    Lib::dagger(tmp);
    if(j) tmp.multiply(-1);
    if(!tmp.equal(diracs[j])) {
      std:: string cmt;
      Lib::smanip::printf(cmt,128," problem with dirac%d+",j);
      Lib::print(diracs[j],printer,cmt);
      Lib::print(tmp,printer,"");
    }
  }}


}
//////////////////////////////////////////////////////////////////////////////
void check_dirac4(
) 
//////////////////////////////////////////////////////////////////////////////
//   dirac4 = dirac0 * dirac1 * dirac2 * dirac3
// We should have :
//   dirac4 * dirac4 = -I
//   dirac4+ = - dirac4
//   { dirac4, dirac(alpha) } = 0
//   dirac4+ * dirac0 = dirac0 * dirac4
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44z dirac4 = Lib::A44z_Dirac0();
  dirac4 *= Lib::A44z_Dirac1();
  dirac4 *= Lib::A44z_Dirac2();
  dirac4 *= Lib::A44z_Dirac3();

 {Lib::A44z tmp = dirac4;
  tmp *= dirac4;
  tmp.multiply(-1);
  Lib::A44z_I id;
  if(!tmp.equal(id)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," problem with dirac4 * dirac4 ");
    Lib::print(tmp,printer,"dirac4 * dirac4");
  }}

 {Lib::A44z tmp = dirac4;
  Lib::dagger(tmp);
  tmp.multiply(-1);
  if(!tmp.equal(dirac4)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," problem with dirac4+ ");
    Lib::print(dirac4,printer,cmt);
    Lib::print(tmp,printer,"dirac4+");
  }}

 {Lib::A44z null;
  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());
  unsigned int diracn = diracs.size();
  for(unsigned int j=0;j<diracn;j++) {
    Lib::A44z ac = Lib::antiCommutator(dirac4,diracs[j]);
    if(!ac.equal(null)) {
      std:: string cmt;
      Lib::smanip::printf(cmt,128," problem with {dirac4,dirac%d}",j);
      Lib::print(ac,printer,cmt);
    }
  }}


 {Lib::A44z dirac0 = Lib::A44z_Dirac0();
  Lib::A44z tmp = dirac4;
  Lib::dagger(tmp);
  Lib::A44z l = tmp * dirac0;
  Lib::A44z r = dirac0 * dirac4;
  if(!l.equal(r)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," problem with dirac4+ * dirac0 ");
    Lib::print(l,printer,cmt);
    Lib::print(r,printer,"r");
  }}

}
//////////////////////////////////////////////////////////////////////////////
void check_dirac0(
) 
//////////////////////////////////////////////////////////////////////////////
//   dirac(alpha)+ * dirac0  = dirac0 * dirac(alpha)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44z dirac0 = Lib::A44z_Dirac0();

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  unsigned int diracn = diracs.size();
  for(unsigned int j=0;j<diracn;j++) {
    Lib::A44z tmp = diracs[j];
    Lib::dagger(tmp);
    Lib::A44z l = tmp * dirac0;
    Lib::A44z r = dirac0 * diracs[j];
    if(!l.equal(r)) {
      std:: string cmt;
      Lib::smanip::printf(cmt,128," problem with dirac(%d)+ * dirac0 ",j);
      Lib::print(l,printer,cmt);
      Lib::print(r,printer,"r");
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void check_rs_commutators(
) 
//////////////////////////////////////////////////////////////////////////////
// [r(j),r(k)] = - epsilon(j,k,l) * r(l)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A33d> rs;
  rs.push_back(Lib::A33d_R1());
  rs.push_back(Lib::A33d_R2());
  rs.push_back(Lib::A33d_R3());

  unsigned int rn = rs.size();
  Lib::A333d_epsilon epsilon;
  for(unsigned int j=0;j<rn;j++) {
    for(unsigned int k=j;k<rn;k++) {
      Lib::A33d cm = Lib::commutator(rs[j],rs[k]);
      Lib::A33d tmp;
      for(unsigned int l=0;l<rn;l++) {
        Lib::A33d tmp1(rs[l]);
        tmp1.multiply(epsilon.value(j,k,l));
        tmp += tmp1;
      }
      tmp.multiply(-1);
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with [r%d,r%d]",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void check_rs_metric(
) 
//////////////////////////////////////////////////////////////////////////////
// trace[r(j),r(k)] = - 2 * delta(j,k)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  std::vector<Lib::A33d> rs;
  rs.push_back(Lib::A33d_R1());
  rs.push_back(Lib::A33d_R2());
  rs.push_back(Lib::A33d_R3());

  unsigned int rn = rs.size();
  Lib::A33d_I id;
  for(unsigned int j=0;j<rn;j++) {
    for(unsigned int k=j;k<rn;k++) {
      Lib::A33d tmp = rs[j] * rs[k];
      double value = tmp.trace();
      if(value!=-2*id.value(j,k)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with trace(r%d*r%d)",j,k);
        out << cmt << " " << value << Lib::endl;
      }
    }
  }

  // Casimir matrix :
 {Lib::A33d result;
  Lib::A33d metric;
  metric.setIdentity();
  metric.multiply(-2.);
  Lib::A33d metric_inv;
  metric.invert(metric_inv);
  for(unsigned int j=0;j<rn;j++) {
    for(unsigned int k=0;k<rn;k++) {
      Lib::A33d tmp = rs[j] * rs[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::A33d casimir(id);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}


}
//////////////////////////////////////////////////////////////////////////////
void check_paulis_metric(
) 
//////////////////////////////////////////////////////////////////////////////
//   trace[pauli(j),pauli(k)] = 2 * delta(j,k)
// and then :
//   trace[i*pauli(j)/2,i*pauli(k)/2] = - 0.5 * delta(j,k)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A22z_Pauli1 pauli1;
  Lib::A22z_Pauli2 pauli2;
  Lib::A22z_Pauli3 pauli3;

  std::vector<Lib::A22z> paulis;
  paulis.push_back(pauli1);
  paulis.push_back(pauli2);
  paulis.push_back(pauli3);

  unsigned int paulin = paulis.size();
  Lib::A33d_I id;
  for(unsigned int j=0;j<paulin;j++) {
    for(unsigned int k=j;k<paulin;k++) {
      Lib::A22z tmp = paulis[j] * paulis[k];
      tmp.multiply(-1./4.);
      z value = tmp.trace();
      if(value!=z(-0.5*id.value(j,k),0)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with trace(r%d*r%d)",j,k);
        out << cmt << " " << Lib::sout(value) << Lib::endl;
      }
    }
  }

  // Casimir matrix :
 {Lib::A22z result;
  Lib::A33d metric;
  metric.setIdentity();
  metric.multiply(-0.5);
  Lib::A33d metric_inv;
  metric.invert(metric_inv);
  for(unsigned int j=0;j<paulin;j++) {
    for(unsigned int k=0;k<paulin;k++) {
      Lib::A22z tmp = paulis[j] * paulis[k];
      tmp.multiply(-1./4.);
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::A22z_I casimir;
  //casimir.multiply(-3./4.);
  casimir.multiply(3./2.);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}
}
//////////////////////////////////////////////////////////////////////////////
void check_taus_metric(
) 
//////////////////////////////////////////////////////////////////////////////
//   trace[tau(j),tau(k)] = 4 * delta(j,k)
// and then (J and tau(j) commutes) :
//   trace[J*tau(j)/2,J*tau(k)/2] = - delta(j,k)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::A44d_T1 T1;
  Lib::A44d_T2 T2;
  Lib::A44d_T3 T3;

  std::vector<Lib::A44d> taus;
  taus.push_back(T1);
  taus.push_back(T2);
  taus.push_back(T3);

  Lib::A33d_I id;
  unsigned int taun = taus.size();
  for(unsigned int j=0;j<taun;j++) {
    for(unsigned int k=j;k<taun;k++) {
      Lib::A44d tmp = taus[j] * taus[k];
      tmp.multiply(-1./4.);
      double value = tmp.trace();
      if(value!=-1.*id.value(j,k)) {
        std:: string cmt;
        Lib::smanip::printf(cmt,128," problem with trace(r%d*r%d)",j,k);
        out << cmt << " " << value << Lib::endl;
      }
    }
  }

  // Casimir matrix :
 {Lib::A44d result;
  Lib::A44d metric;
  metric.setIdentity();
  metric.multiply(-1.);
  Lib::A44d metric_inv;
  metric.invert(metric_inv);
  for(unsigned int j=0;j<taun;j++) {
    for(unsigned int k=0;k<taun;k++) {
      Lib::A44d tmp = taus[j] * taus[k];
      tmp.multiply(-1./4.);
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::A44d casimir;
  casimir.setIdentity();
  casimir.multiply(3./4.);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

}
//////////////////////////////////////////////////////////////////////////////
bool check_En_csts(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  for(unsigned int order=2;order<7;order++) {

  unsigned int number = order * (order-1)/2;

  Lib::Array<double> csts;
  if(!Lib::ADDd_E::groupConstants(order,csts)) {
    out << "check_En_csts :"
        << " problem getting group constants."
        << Lib::endl;
    return false;
  }

/*
  std:: string cmt;
  Lib::smanip::printf(cmt,128,"For E%d (%d) :",order,number);
  printer.write(cmt);
  printer.write("\n");

  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j;k<number;k++) {
      std::vector<unsigned int> is(3);
      is[0] = j;
      is[1] = k;
      for(unsigned int l=0;l<number;l++) {
        is[2] = l;
        double value;
        if(!csts.value(is,value)) {
          out << "check_En_csts :"
              << " can't get value."
              << Lib::endl;
          return false;
        }
        if(value==1) {
          std:: string cmt;
          Lib::smanip::printf(cmt,128," [%d,%d] =  %d",j,k,l);
          printer.write(cmt);
          printer.write("\n");
          break;
        } else if(value==-1) {
          std:: string cmt;
          Lib::smanip::printf(cmt,128," [%d,%d] = -%d",j,k,l);
          printer.write(cmt);
          printer.write("\n");
          break;
        } else if(value!=0) {
          std:: string cmt;
          Lib::smanip::printf(cmt,128," [%d,%d] = %g * %d ???",j,k,value,l);
          printer.write(cmt);
          printer.write("\n");
          break;
        }
      }
    }
  }
*/

  // order 2 ->  0*I(1)
  // order 3 -> -2*I(3)
  // order 4 -> -4*I(6)
  // order 5 -> -6*I(10)
  // order 6 -> -8*I(15)
  // order 7 -> -10*I(21)
  // order n -> -(2*n-4))*I(n*(n-1)/2)

 {Lib::ADD<double> metric(number);
  if(!Lib::ADDd_E::cartanMetric(order,metric)) {
    out << "check_En_csts :"
        << " problem getting cartan metric."
        << Lib::endl;
    return false;
  }
  std::string cmt;
  Lib::smanip::printf(cmt,128,"For E%d (%d), cartan metric :",order,number);
  Lib::print(metric,printer,cmt);}

 {Lib::ADD<double> metric(number);
  if(!Lib::ADDd_E::metric(order,metric)) {
    out << "check_En_csts :"
        << " problem getting metric."
        << Lib::endl;
    return false;
  }
  std::string cmt;
  Lib::smanip::printf(cmt,128,"For E%d (%d), metric :",order,number);
  Lib::print(metric,printer,cmt);}

 {Lib::ADD<double> casimir(number);
  if(!Lib::ADDd_E::casimir(order,casimir)) {
    out << "check_En_csts :"
        << " problem getting casimir metric."
        << Lib::endl;
    return false;
  }
  std::string cmt;
  Lib::smanip::printf(cmt,128,"For E%d (%d), casimir metric :",order,number);
  Lib::print(casimir,printer,cmt);}

  } //end for order.

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_rotation3D_vector_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Vector representation of the rotation3D group :
//   VecRep_k = E3_k
//   k in [0,3[
// We must have :
//  [VecRep_j,VecRep_k] = Rotation_Group_const(j,k,l) * VecRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 3;
  unsigned int number = order * (order-1)/2;

  //std::vector<Lib::A44d> reps;

  std::vector< Lib::ADD<double> > reps;

  for(unsigned int index=0;index<number;index++) {
    Lib::ADDd_E En(order,index);
    reps.push_back(En);
  }

/*
 {for(unsigned int index=0;index<number;index++) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," VecRep_%d :",index);
    Lib::print(reps[index],printer,cmt);
  }}
*/

  Lib::A333d_epsilon csts;

  // Check the reps commutators :
 {for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::ADD<double> cm = Lib::commutator(reps[j],reps[k]);
      //std:: string cmt;
      //Lib::smanip::printf(cmt,128," [VecRep_%d,VecRep_%d] :",j,k);
      //Lib::print(cm,printer,cmt);

      Lib::A33d tmp;
      for(unsigned int l=0;l<number;l++) {
        Lib::ADD<double> tmp1(reps[l]);
        tmp1.multiply(csts.value(j,k,l));
        tmp += tmp1;
      }
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf
          (cmt,128," Problem with [VecRep_%d,VecRep_%d] :",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }
    }
  }}

  // metric of the rep :
  //   Trace(rep[j],rep[k])
  // Seems to be :
  //  -2  0  0
  //   0 -2  0
  //   0  0 -2
 {Lib::ADD<double> metric(number);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = reps[j] * reps[k];
      metric.setValue(j,k,mx.trace());
    }
  }
  Lib::A33d_I m;
  m.multiply(-2.);
  if(!m.equal(metric)) {
    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Problem with rotation3D vec rep metric :");
    Lib::print(metric,printer,cmt);
    Lib::print(m,printer,"expected :");
  }}

  // Casimir matrix :
 {Lib::A33d result;
  Lib::A33d_I metric;
  metric.multiply(-2.);
  Lib::A33d metric_inv;
  metric.invert(metric_inv);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> tmp = reps[j] * reps[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::ADD<double> casimir(order);
  casimir.setIdentity();
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_lorentz_vector_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Vector representation of the Lorentz group :
//   VecRep_k = eta * E4_k
//   k in [0,6[
// We must have :
//  [VecRep_j,VecRep_k] = Lorentz_Group_const(j,k,l) * VecRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 4;
  unsigned int number = order * (order-1)/2;

  std::vector< Lib::ADD<double> > reps;

  Lib::A44d_eta eta;

  for(unsigned int index=0;index<number;index++) {
    Lib::ADDd_E En(order,index);
    Lib::ADD<double> result(order);
    result = eta * En;
    reps.push_back(result);
  }

/*
 {for(unsigned int index=0;index<number;index++) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," VecRep_%d :",index);
    Lib::print(reps[index],printer,cmt);
  }}
*/

  Lib::A666d_LorentzGroupConstants csts;

  // Check the reps commutators :
 {for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::ADD<double> cm = Lib::commutator(reps[j],reps[k]);
      //std:: string cmt;
      //Lib::smanip::printf(cmt,128," [VecRep_%d,VecRep_%d] :",j,k);
      //Lib::print(cm,printer,cmt);

      Lib::A44d tmp;
      for(unsigned int l=0;l<number;l++) {
        Lib::ADD<double> tmp1(reps[l]);
        tmp1.multiply(csts.value(j,k,l));
        tmp += tmp1;
      }
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf
          (cmt,128," Problem with [VecRep_%d,VecRep_%d] :",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }

    }
  }}

 //{Lib::A66d_LorentzCartanMetric m;
  //Lib::print(m,printer,"Lorentz cartan metric :");}

  // metric of the rep :
  //   Trace(rep[j],rep[k])
  // Seems to be :
  //   2  0  0  0  0  0
  //   0  2  0  0  0  0
  //   0  0 -2  0  0  0
  //   0  0  0  2  0  0
  //   0  0  0  0 -2  0
  //   0  0  0  0  0 -2
 {Lib::ADD<double> metric(number);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> mx = reps[j] * reps[k];
      metric.setValue(j,k,mx.trace());
    }
  }
  Lib::A66d_LorentzVecMetric m;
  if(!m.equal(metric)) {
    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Problem with lorentz vec rep metric :");
    Lib::print(metric,printer,cmt);
    Lib::print(m,printer,"expected :");
  }}

  // Casimir matrix :
 {Lib::A44d result;
  Lib::A66d_LorentzVecMetric metric;
  Lib::ADD<double> metric_inv(number);
  metric.invert(metric_inv);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<double> tmp = reps[j] * reps[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::ADD<double> casimir(order);
  casimir.setIdentity();
  casimir.multiply(1.5);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_lorentz_dirac_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Spinor representation of the Lorentz group :
//   SpinRep_k = Sum(alpha,beta) of [dirac(alpha),dirac(beta)]*E4_k(alpha,beta)
//   k in [0,6[
// We must have :
//  [SpinRep_j,SpinRep_k] = Lorentz_Group_const(j,k,l) * SpinRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 4;
  unsigned int number = order * (order-1)/2;

  std::vector<Lib::A44z> reps;

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  unsigned int diracn = diracs.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ADDd_E En(order,index);
    Lib::A44z result;
    for(unsigned int j=0;j<diracn;j++) {
      for(unsigned int k=j+1;k<diracn;k++) {
        Lib::A44z cm = Lib::commutator(diracs[j],diracs[k]);
        cm.multiply(En.value(j,k));
        result += cm;
      }
    }
    result.multiply(2);
    result.divide(8);
    reps.push_back(result);
  }

/*
 {for(unsigned int index=0;index<number;index++) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," SpinRep_%d :",index);
    Lib::print(reps[index],printer,cmt);
  }}
*/

  Lib::A666d_LorentzGroupConstants csts;

  // Check the reps commutators :
 {for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::A44z cm = Lib::commutator(reps[j],reps[k]);
      //std:: string cmt;
      //Lib::smanip::printf(cmt,128," [SpinRep_%d,SpinRep_%d] :",j,k);
      //Lib::print(cm,printer,cmt);

      Lib::A44z tmp;
      for(unsigned int l=0;l<number;l++) {
        Lib::A44z tmp1(reps[l]);
        tmp1.multiply(csts.value(j,k,l));
        tmp += tmp1;
      }
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf
          (cmt,128," Problem with [SpinRep_%d,SpinRep_%d] :",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }

    }
  }}

  // Metric of the rep :
  //   Trace(rep[j],rep[k])
  // Seems to be :
  //   1  0  0  0  0  0
  //   0  1  0  0  0  0
  //   0  0 -1  0  0  0
  //   0  0  0  1  0  0
  //   0  0  0  0 -1  0
  //   0  0  0  0  0 -1
 {Lib::ADD<z> metric(number);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<z> mx = reps[j] * reps[k];
      metric.setValue(j,k,mx.trace());
    }
  }
  Lib::A66z_LorentzDiracMetric m;
  if(!m.equal(metric)) {
    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Problem with lorentz dirac rep metric :");
    Lib::print(metric,printer,cmt);
    Lib::print(m,printer,"expected :");
  }}

  // Casimir matrix :
 {Lib::A44z result;
  Lib::A66z_LorentzDiracMetric metric;
  Lib::ADD<z> metric_inv(6);
  metric.invert(metric_inv);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::A44z tmp = reps[j] * reps[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::A44z_I casimir;
  casimir.multiply(1.5);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_lorentz_weyl_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Weyl spinor representation of the Lorentz group :
//   SpinRep_k = (pauli_1/2,pauli_2/2,pauli_3/2,
//                i*pauli_1/2,i*pauli_2/2,i*pauli_3/2)
//   k in [0,6[
// We must have :
//  [SpinRep_j,SpinRep_k] = Lorentz_Group_const(j,k,l) * SpinRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  unsigned int number = 6;

  z zi(0,1);

  Lib::A22z_Pauli1 pauli1;
  Lib::A22z_Pauli2 pauli2;
  Lib::A22z_Pauli3 pauli3;

  Lib::A666d_LorentzGroupConstants csts;

  std::vector<Lib::A22z> reps;
 {Lib::A22z tmp(pauli1);
  tmp.multiply(0.5);  
  reps.push_back(tmp);}
 {Lib::A22z tmp(pauli2);
  tmp.multiply(0.5);  
  reps.push_back(tmp);}
 {Lib::A22z tmp(pauli3);
  tmp.multiply(0.5);  
  tmp.multiply(zi);  
  reps.push_back(tmp);}

 {Lib::A22z tmp(pauli3);
  tmp.multiply(0.5);  
  reps.push_back(tmp);}
 {Lib::A22z tmp(pauli2);
  tmp.multiply(-0.5);  
  tmp.multiply(zi);  
  reps.push_back(tmp);}
 {Lib::A22z tmp(pauli1);
  tmp.multiply(0.5);  
  tmp.multiply(zi);  
  reps.push_back(tmp);}

  // Check the reps commutators :
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::A22z cm = Lib::commutator(reps[j],reps[k]);
      //std:: string cmt;
      //Lib::smanip::printf(cmt,128," [SpinRep_%d,SpinRep_%d] :",j,k);
      //Lib::print(cm,printer,cmt);
  
      Lib::A22z tmp;
      for(unsigned int l=0;l<number;l++) {
        Lib::A22z tmp1(reps[l]);
        tmp1.multiply(csts.value(j,k,l));
        tmp += tmp1;
      }
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf
          (cmt,128," Problem with [SpinRep_%d,SpinRep_%d] :",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }  
    }
  }

/*
  Lib::ADD<z> metric(number);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<z> mx = reps[j] * reps[k];
      metric.setValue(j,k,mx.trace());
    }
  }
 {Lib::A66z_LorentzWeylMetric m;
  if(!m.equal(metric)) {
    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Problem with lorentz weyl rep metric :");
    Lib::print(metric,printer,cmt);
    Lib::print(m,printer,"expected :");
  }}
  //FIXME : rep metric not proportional to a diagonal matrix !
  //FIXME : does it mean that the rep is reductible ?
  //FIXME : rep metric not invertible !!! Is it ok ?
*/

  // Casimir matrix :
 {Lib::A22z result;
  Lib::ADD<z> metric(number);
  metric.setValue(0,0, 1.);
  metric.setValue(1,1, 1.);
  metric.setValue(2,2,-1.);
  metric.setValue(3,3, 1.);
  metric.setValue(4,4,-1.);
  metric.setValue(5,5,-1.);
  //metric.multiply(4.);

  Lib::ADD<z> metric_inv(number);
  if(!metric.invert(metric_inv)) {
    out << "check_lorentz_weyl_rep :"
        << " can't invert metric"
        << Lib::endl;
    return false;
  }
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::A22z tmp = reps[j] * reps[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  Lib::A22z_I casimir;
  casimir.multiply(3./2.);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_poincare_op_rep(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  //Lib::Out out(printer);

  for(unsigned int order=2;order<=5;order++) {

    Lib::ADDd_PoincareCartanMetric m(order);

    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Poincare cartan metric for order %d :",order);
    Lib::print(m,printer,cmt);

  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_delta_epsilon_33(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  unsigned int order = 3;
  Lib::A33d_I delta;

  Lib::A333d_epsilon epsilon;

 {for(unsigned int alpha=0;alpha<order;alpha++) {
  for(unsigned int m=0;m<order;m++) {
  for(unsigned int n=0;n<order;n++) {
  for(unsigned int a=0;a<order;a++) {
  for(unsigned int b=0;b<order;b++) {
    double left = 0;
    left += delta.value(m,a)*epsilon.value(alpha,n,b);
    left -= delta.value(n,a)*epsilon.value(alpha,m,b);
    left -= delta.value(n,b)*epsilon.value(alpha,a,m);
    left += delta.value(m,b)*epsilon.value(alpha,a,n);
    double right = 0;
    right += delta.value(m,alpha)*epsilon.value(a,n,b);
    right -= delta.value(n,alpha)*epsilon.value(a,m,b);
    if(left!=right) {
      printer.write("check_delta_epsilon_33 : fails\n");
      return false;
    }
    // One single right term does not seem to be a combination
    //   of two of the left.
  }}}}}}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_J_mu_nu_33(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  unsigned int order = 3;
  typedef double Type;
  typedef Lib::A33d Lib_Matrix;

  Lib_Matrix eta;
  eta.setIdentity();

  Lib_Matrix delta;
  delta.setIdentity();

  //    a                  b                b
  // J      = eta   * delta   - eta  * delta
  //  mn b       ma        n       na       m
  std::vector<Lib_Matrix> Js;
 {for(unsigned int m=0;m<order;m++) {
  for(unsigned int n=0;n<order;n++) {
    Lib_Matrix J;
    for(unsigned int a=0;a<order;a++) {
      for(unsigned int b=0;b<order;b++) {
        Type value = eta.value(m,a)*delta.value(b,n)
                   - eta.value(n,a)*delta.value(b,m);
        J.setValue(a,b,value);
      }
    }
    Js.push_back(J);
  }}}

  // Check Ps such that :
  // [J_m_n,P_r] = -eta_m_r * P_n + eta_n_r*P_m
  std::vector<Lib_Matrix> Ps(3);

  double p0_01 = 0;
  double p0_02 = 0;
  double p0_12 = 3.14;
  Ps[0].setValue(0,1,p0_01);
  Ps[0].setValue(0,2,p0_02);
  Ps[0].setValue(1,2,p0_12);
  Ps[0].setValue(1,0,-p0_01);
  Ps[0].setValue(2,0,-p0_02);
  Ps[0].setValue(2,1,-p0_12);

  double p1_01 = 0;
  double p1_02 = -p0_12;
  double p1_12 = 0;
  Ps[1].setValue(0,1,p1_01);
  Ps[1].setValue(0,2,p1_02);
  Ps[1].setValue(1,2,p1_12);
  Ps[1].setValue(1,0,-p1_01);
  Ps[1].setValue(2,0,-p1_02);
  Ps[1].setValue(2,1,-p1_12);

  double p2_01 = p0_12;
  double p2_02 = 0;
  double p2_12 = 0;
  Ps[2].setValue(0,1,p2_01);
  Ps[2].setValue(0,2,p2_02);
  Ps[2].setValue(1,2,p2_12);
  Ps[2].setValue(1,0,-p2_01);
  Ps[2].setValue(2,0,-p2_02);
  Ps[2].setValue(2,1,-p2_12);

  bool ok = true;
  for(unsigned int m=0;m<order;m++) {
    for(unsigned int n=m+1;n<order;n++) {
      for(unsigned int r=0;r<order;r++) {
        Lib_Matrix cmt = Lib::commutator(Js[n+m*order],Ps[r]);
        Lib_Matrix mtx = Ps[n];
        mtx.multiply(-eta.value(m,r));
        Lib_Matrix tmp = Ps[m];
        tmp.multiply(eta.value(n,r));
        mtx += tmp;
        if(!cmt.equal(mtx)) {
          std:: string scmt;
          Lib::smanip::printf(scmt,128," mu %d, nu %d, P %d :",m,n,r);
          Lib::print(cmt,printer,scmt);
          Lib::print(mtx,printer,"mtx");
          Lib::print(Js[n+m*order],printer,"J");
          Lib::print(Ps[r],printer,"P");
          ok = false;
          break;
        }
      }
      if(!ok) break;
    }
    if(!ok) break;
  }

  if(ok) {
    //printer.write("check_J_mu_nu_33 : ouai\n");
    //for(unsigned int r=0;r<order;r++) {
    //  std:: string cmt;
    //  Lib::smanip::printf(cmt,128," P %d :",r);
    //  Lib::print(Ps[r],printer,cmt);
    //}
  } else {
    printer.write("check_J_mu_nu_33 : fails\n");
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_kronecker(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  for(unsigned int order = 1;order<=4;order++) {

  out << "check_kronecker : " << order << Lib::endl;

  Lib::Kronecker<double> k(order);
  const std::vector<double>& vec = k.vector();
  Lib::Array<double>::UInts is(order);
  unsigned int index = 0;
  Lib::Array<double>::ConsVecIt it = vec.begin();
  for(;it!=vec.end();++it,index++) {
    double value = (*it);
    if(value==0) continue;
    if(!k.indices(index,is)) return false;
    for(unsigned iaxis=0;iaxis<order;iaxis++) out << " " << is[iaxis];
    out << " = " << value << Lib::endl;
  }

  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_gb_eta_dirac_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Spinor representation of the GB group :
//   SpinRep_k = 
//     Sum(alpha,beta) of [dirac(alpha),dirac(beta)]*(eta*E4_k)(alpha,beta)
//   k in [0,5]
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 4;
  unsigned int number = order * (order-1)/2;

  std::vector<Lib::A44z> reps;

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  Lib::A44d_eta eta;

  unsigned int diracn = diracs.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ADDd_E En(order,index);
    Lib::ADD<double> tmp(order);
    tmp.setIdentity();
    tmp.multiply(-1);
    tmp.setValue(1,1,1);
    tmp = tmp * En;

    Lib::A44z result;
    for(unsigned int j=0;j<diracn;j++) {
      for(unsigned int k=j+1;k<diracn;k++) {
        Lib::A44z cm = Lib::commutator(diracs[j],diracs[k]);
        cm.multiply(tmp.value(j,k));
        result += cm;
      }
    }
    reps.push_back(result);
  }

  //reps.push_back(Lib::A44z_Dirac0());
  //reps.push_back(Lib::A44z_Dirac1());
  //reps.push_back(Lib::A44z_Dirac2());
  //reps.push_back(Lib::A44z_Dirac3());

  number = reps.size();
 {for(unsigned int index=0;index<number;index++) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," SpinRep_%d :",index);
    Lib::print(reps[index],printer,cmt);
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_gb_dirac_rep(
) 
//////////////////////////////////////////////////////////////////////////////
// Spinor representation of the GB group :
//   SpinRep_k = Sum(alpha,beta) of [dirac(alpha),dirac(beta)]*E4_k(alpha,beta)
//   k in [0,5]
//   SpinRep_k = Diracs
//   k in [6,7,8,9]
// We must have :
//  [SpinRep_j,SpinRep_k] = GB_Group_const(j,k,l) * SpinRep_l
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;

  unsigned int order = 4;
  unsigned int number = order * (order-1)/2;

  std::vector<Lib::A44z> reps;

  std::vector<Lib::A44z> diracs;
  diracs.push_back(Lib::A44z_Dirac0());
  diracs.push_back(Lib::A44z_Dirac1());
  diracs.push_back(Lib::A44z_Dirac2());
  diracs.push_back(Lib::A44z_Dirac3());

  unsigned int diracn = diracs.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ADDd_E En(order,index);
    Lib::A44z result;
    for(unsigned int j=0;j<diracn;j++) {
      for(unsigned int k=j+1;k<diracn;k++) {
        Lib::A44z cm = Lib::commutator(diracs[j],diracs[k]);
        cm.multiply(En.value(j,k));
        result += cm;
      }
    }
    result.multiply(2);
    result.divide(8);
    reps.push_back(result);
  }

  reps.push_back(Lib::A44z_Dirac0());
  reps.push_back(Lib::A44z_Dirac1());
  reps.push_back(Lib::A44z_Dirac2());
  reps.push_back(Lib::A44z_Dirac3());

  number = reps.size();
/*
 {for(unsigned int index=0;index<number;index++) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," SpinRep_%d :",index);
    Lib::print(reps[index],printer,cmt);
  }}
*/

  //Lib::A666d_LorentzGroupConstants csts;

/*
  // Check the reps commutators :
 {for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=j+1;k<number;k++) {
      Lib::A44z cm = Lib::commutator(reps[j],reps[k]);
      //std:: string cmt;
      //Lib::smanip::printf(cmt,128," [SpinRep_%d,SpinRep_%d] :",j,k);
      //Lib::print(cm,printer,cmt);

      Lib::A44z tmp;
      for(unsigned int l=0;l<number;l++) {
        Lib::A44z tmp1(reps[l]);
        tmp1.multiply(csts.value(j,k,l));
        tmp += tmp1;
      }
      if(!cm.equal(tmp)) {
        std:: string cmt;
        Lib::smanip::printf
          (cmt,128," Problem with [SpinRep_%d,SpinRep_%d] :",j,k);
        Lib::print(cm,printer,cmt);
        Lib::print(tmp,printer,"tmp");
      }

    }
  }}
*/

  // Metric of the rep :
  //   Trace(rep[j],rep[k])
  // Seems to be :
  //   1  0  0  0  0  0  0  0  0  0 
  //   0  1  0  0  0  0  0  0  0  0
  //   0  0 -1  0  0  0  0  0  0  0
  //   0  0  0  1  0  0  0  0  0  0
  //   0  0  0  0 -1  0  0  0  0  0
  //   0  0  0  0  0 -1  0  0  0  0
  //   0  0  0  0  0  0  4  0  0  0
  //   0  0  0  0  0  0  0 -4  0  0
  //   0  0  0  0  0  0  0  0 -4  0
  //   0  0  0  0  0  0  0  0  0 -4
  Lib::ADD<z> metric(number);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::ADD<z> mx = reps[j] * reps[k];
      metric.setValue(j,k,mx.trace());
    }
  }
  //std:: string cmt;
  //Lib::smanip::printf(cmt,128," Rep metric :");
  //Lib::print(metric,printer,cmt);

/*
  Lib::A66z_LorentzDiracMetric m;
  if(!m.equal(metric)) {
    std:: string cmt;
    Lib::smanip::printf
      (cmt,128," Problem with lorentz dirac rep metric :");
    Lib::print(metric,printer,cmt);
    Lib::print(m,printer,"expected :");
  }}
*/

  // Casimir matrix :
 {Lib::A44z result;
  //Lib::A66z_LorentzDiracMetric metric;
  Lib::ADD<z> metric_inv(6);
  metric.invert(metric_inv);
  for(unsigned int j=0;j<number;j++) {
    for(unsigned int k=0;k<number;k++) {
      Lib::A44z tmp = reps[j] * reps[k];
      tmp.multiply(metric_inv.value(j,k));
      result += tmp;
    }
  }
  //std:: string cmt;
  //Lib::smanip::printf(cmt,128," Casimir matrix :");
  //Lib::print(result,printer,cmt);
  Lib::A44z_I casimir;
  casimir.multiply(2.5);
  if(!result.equal(casimir)) {
    std:: string cmt;
    Lib::smanip::printf(cmt,128," Problem with Casimir :");
    Lib::print(result,printer,cmt);
    Lib::print(casimir,printer,"expected :");
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool check_determinant_with_kronecker(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  typedef z Type;
  typedef Lib::ADD<Type> Lib_Matrix;

  for(unsigned int order = 1;order<=6;order++) {

    Lib_Matrix mtx(order);
    mtx.setIdentity();
    mtx.multiply(2.);

    Type det;
    if(!Lib::determinant<Type>(mtx,det)) return false;

    Type mdet = mtx.determinant();
    if(det!=mdet) {
      out << "check_determinant_with_kronecker :"
          << " problem with order " << order
          << ". Found " << det << " whilst " << mdet << " expected."
          << Lib::endl;
    }

  }

 {Lib::A44d_T1 mtx;
  double det;
  if(!Lib::determinant<double>(mtx,det)) return false;
  double mdet = mtx.determinant();
  if(det!=mdet) {
    out << "check_determinant_with_kronecker :"
        << " problem with T1."
        << " Found " << det << " whilst " << mdet << " expected."
        << Lib::endl;
  }}

 {Lib::A44d_T2 mtx;
  double det;
  if(!Lib::determinant<double>(mtx,det)) return false;
  double mdet = mtx.determinant();
  if(det!=mdet) {
    out << "check_determinant_with_kronecker :"
        << " problem with T2."
        << " Found " << det << " whilst " << mdet << " expected."
        << Lib::endl;
  }}

 {Lib::A44d_T3 mtx;
  double det;
  if(!Lib::determinant<double>(mtx,det)) return false;
  double mdet = mtx.determinant();
  if(det!=mdet) {
    out << "check_determinant_with_kronecker :"
        << " problem with T3."
        << " Found " << det << " whilst " << mdet << " expected."
        << Lib::endl;
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool search_mtx_33(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  unsigned int order = 3;
  typedef z Type;
  typedef Lib::ADD<Type> Lib_Matrix;

  std::vector<Lib_Matrix> rs;
  rs.push_back(Lib::tocomplex<double>(Lib::A33d_R1()));
  rs.push_back(Lib::tocomplex<double>(Lib::A33d_R2()));
  rs.push_back(Lib::tocomplex<double>(Lib::A33d_R3()));

  unsigned int nspace = order*order;
  std::vector< std::vector<Type> > space(nspace);
  unsigned int nval = 5;
 {std::vector<Type> vals(nval);
  vals[0] = 0;
  vals[1] = 1;
  vals[2] = -1;
  z zi(0,1);
  vals[3] = zi;
  vals[4] = -zi;
  for(unsigned int ispace=0;ispace<nspace;ispace++) {
    space[ispace] = vals;
  }}

  std::vector<Type> slice(nspace,0);

  printer.write("search_mtx_33 : start search...\n");

  for(unsigned int i_0=0;i_0<nval;i_0++) {
  slice[0] = space[0][i_0];
  for(unsigned int i_1=0;i_1<nval;i_1++) {
  slice[1] = space[1][i_1];
  for(unsigned int i_2=0;i_2<nval;i_2++) {
  slice[2] = space[2][i_2];
  for(unsigned int i_3=0;i_3<nval;i_3++) {
  slice[3] = space[3][i_3];
  for(unsigned int i_4=0;i_4<nval;i_4++) {
  slice[4] = space[4][i_4];
  for(unsigned int i_5=0;i_5<nval;i_5++) {
  slice[5] = space[5][i_5];
  for(unsigned int i_6=0;i_6<nval;i_6++) {
  slice[6] = space[6][i_6];
  for(unsigned int i_7=0;i_7<nval;i_7++) {
  slice[7] = space[7][i_7];
  for(unsigned int i_8=0;i_8<nval;i_8++) {
  slice[8] = space[8][i_8];

  Lib_Matrix M(order);

 {for(unsigned int a=0;a<order;a++) {
    for(unsigned int b=0;b<order;b++) {
      unsigned int offset = a+b*order;
      M.setValue(a,b,slice[offset]);
    }
  }}

  bool ok = true;
 {for(unsigned int i=0;i<order;i++) {
    Lib_Matrix result = M*rs[i]*M;
    result.transpose(); //Without transpose, we correctly find I and -I.
    if(!result.equal(rs[i])) {
      ok = false;
      break;
    }
  }}

  if(ok) {
    printer.write("search_mtx_33 : ouai\n");
    std:: string cmt;
    Lib::smanip::printf(cmt,128," M :");
    Lib::print(M,printer,cmt);
  //} else {
    //printer.write("search_mtx_33 : not found\n");
  }

  }}}}}}}}}

  printer.write("search_mtx_33 : end search.\n");

  return true;
}

//#include <Lib/Time.h>
//////////////////////////////////////////////////////////////////////////////
int main(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);
  out << "Lib_algebra :" 
      << " 28." 
      << Lib::endl;

  check_eta_J(); 

  check_paulis(); 
  check_paulis_anticommutators();

  check_weyl(); 
  check_weyl_SUSY(); 
  check_weyl_SUSY_2(); 
  check_dirac_SUSY_dirac();
  //check_dirac_lorentz_SUSY_generators();
  check_weyl_generators();
  check_weyl_generators_anti();

  check_vector_lorentz_generators();

  check_diracs_anticommutators();
  check_dirac_commutator_commutator(); 
  check_dirac_dagger(); 
  check_dirac0(); 
  check_dirac4(); 
  check_dirac_lorentz_generators();

  check_taus_anticommutators();
  check_taus_J_commutators(); 
  check_taus_transpose(); 

  // 3D rotation group :
  check_rs_commutators();
  check_paulis_commutators();
  check_taus_commutators(); 

  check_rs_metric(); 
  check_paulis_metric(); 
  check_taus_metric(); 

  //check_En_csts(); 

  // Check some group representations (Done with the general En matrices) :
  check_rotation3D_vector_rep();
  check_lorentz_vector_rep(); 
  check_lorentz_dirac_rep(); 
  check_lorentz_weyl_rep(); 
  //check_poincare_op_rep(); 
  check_delta_epsilon_33(); 
  check_J_mu_nu_33();
  //check_gb_eta_dirac_rep(); 
  check_gb_dirac_rep(); 
  //check_kronecker(); 
  check_determinant_with_kronecker(); 
  //search_mtx_33(); 

  //Lib::Time start = Lib::Time::now();
  //Lib::Time elapsed = Lib::Time::elapsed(start);
  //out << " elapsed " << elapsed.seconds() << " sec." << Lib::endl;

  return 0;
}
